from astropy.table import Table
from astropy.io import fits
from scipy.interpolate import griddata

import pynrc, webbpsf, os
#inst = webbpsf.NIRCam()

outdir = pynrc.conf.path

data_dir = webbpsf.utils.get_webbpsf_data_path() + '/'
zernike_file = data_dir + 'si_zernikes_isim_cv3.fits'

if not os.path.exists(zernike_file):
    print('Zernike file does not exist:')
    print('  {}'.format(zernike_file))
else:
    ztable_full = Table.read(zernike_file)

    keys = np.array(ztable_full.keys())
    ind_z = ['Zernike' in k for k in keys]
    zkeys = keys[ind_z]

    for mod in ['SW', 'LW', 'SWA', 'LWA', 'SWB', 'LWB']:
        ind_nrc = ['NIRCam'+mod in row['instrument'] for row in ztable_full]

        # Grab V2/V3 coordinates
        # In units of arcsec
        v2 = ztable_full[ind_nrc]['V2']
        v3 = ztable_full[ind_nrc]['V3']

        # Create finer mesh grid
        dstep = 1. / 60. # 1" steps
        xgrid = np.arange(v2.min()-dstep/2, v2.max()+dstep/2, dstep)
        ygrid = np.arange(v3.min()-dstep/2, v3.max()+dstep/2, dstep)
        X, Y = np.meshgrid(xgrid,ygrid)

        # Create a Zernike cube
        zcube = []
        for k in zkeys:
    
            z = ztable_full[ind_nrc][k]

            # There will be some NaNs along the outer borders
            zgrid = griddata((v2, v3), z, (X, Y), method='cubic')
            ind_nan = np.isnan(zgrid)

            # Replace NaNs with nearest neighbors
            zgrid2 = griddata((X[~ind_nan], Y[~ind_nan]), zgrid[~ind_nan], (X,Y), method='nearest')
            zgrid[ind_nan] = zgrid2[ind_nan]

            #ind_min = zgrid<z.min()
            #ind_max = zgrid>z.max()
            #zgrid[ind_min] = z.min()
            #zgrid[ind_max] = z.max()
    
            zcube.append(zgrid)

        zcube = np.array(zcube)

        hdu = fits.PrimaryHDU(zcube)
        hdr = hdu.header

        hdr['units'] = 'meters'
        hdr['xunits'] = 'Arcsec'
        hdr['xmin'] = X.min()
        hdr['xmax'] = X.max()
        hdr['xdel'] = dstep
        hdr['yunits'] = 'Arcsec'
        hdr['ymin'] = Y.min()
        hdr['ymax'] = Y.max()
        hdr['ydel'] = dstep
    
        #hdr['wave'] = 

        hdr['comment'] = 'X and Y values correspond to V2 and V3 coordinates (arcsec).'
        hdr['comment'] = 'Slices in the cube correspond to Zernikes 1 to 36.'
        hdr['comment'] = 'Zernike values calculated using 2D cubic interpolation.'

        fname = 'NIRCam{}_zernikes_isim_cv3.fits'.format(mod)
        hdu.writeto(outdir + fname, overwrite=True)


    #plt.clf()
    #plt.contourf(xgrid,ygrid,zcube[i],20)
    #plt.scatter(v2,v3,marker='o',c='k',s=5)
    #plt.xlim([X.min(),X.max()])
    #plt.ylim([Y.min(),Y.max()])
    #plt.axes().set_aspect('equal', 'datalim')




from __future__ import print_function, division

# Import the usual libraries
import numpy as np

import pynrc
from pynrc import nrc_utils
from pynrc.nrc_utils import webbpsf, poppy, offset_bar

import astropy.io.fits as fits
import multiprocessing as mp

pynrc.setup_logging('WARNING', verbose=False)

from poppy import zernike
from poppy.optics import MultiHexagonAperture
#from poppy.utils import pad_to_size
from pynrc.maths.image_manip import pad_or_cut_to_size

class OPD_extract(object):

    """Decompose OPD to Zernike/Hexike coefficients
    
    For a given JWST OPD image and header, extract the Zernike/Hexike 
    components for the overall pupil and each mirror segment. Makes
    use of functions in webbpsf.
    
    Parameters
    ----------
    opd : ndarray
        OPD image map.
    header : obj
        Header extracted from OPD FITS files.
    seg_terms : int
        Number of Hexike terms to fit to each mirror segment.
    verbose : bool
        Print information as fittings are performed.
    """

    def __init__(self, opd, header, seg_terms=30, verbose=False):

        self.opd = opd
        self.npix = opd.shape[0]
        self.pix_scale = header['PUPLSCAL'] # pupil scale in meters/pixel

        # Create a list of segment masks
        self.nseg = 18
        # Flat to flat is 1.32 meters with a 3 mm air gain between mirrors
        # Supposedly there's a 15 mm effective gap due to polishing
        f2f = 1.308
        gap = 0.015

        # Analytic pupil mask for all segments
        # To work with the OPD Rev V masks, use npix=1016 and pad to 1024
        self._mask_pupil = MultiHexagonAperture(rings=2, flattoflat=f2f, gap=gap)

        # Individual segments in a list
        self._mask_segs = []
        for i in np.arange(self.nseg)+1:
            self._mask_segs.append(MultiHexagonAperture(rings=2, flattoflat=f2f, gap=gap, segmentlist=[i]))

        # Get the x/y positions 
        self.segment_positions = [self._get_seg_xy(i) for i in range(self.nseg)]

        self.basis_zernike = zernike.zernike_basis_faster
        self.basis_hexike  = zernike.hexike_basis

        if verbose: print('Fitting Zernike coefficients for entire pupil...')
        self._coeff_pupil = self._get_coeff_pupil()
        if verbose: print('Constructing OPD from entire pupil coefficients...')
        self.opd_new_pupil = self._get_opd_pupil()
        self.opd_diff_pupil = self.opd - self.opd_new_pupil

        if verbose: print('Fitting Hexike coefficients for each segment...')
        self._coeff_segs = self._get_coeff_segs(nterms=seg_terms)
        if verbose: print('Constructing OPD for each segment...')
        self.opd_new_segs = self._get_opd_new_segs()

        if verbose: print('Finished.')

    @property
    def mask_pupil(self):
        """Tricontagon mask"""
        outmask, _ = self._sample_mask(self._mask_pupil)
        return outmask

    @property
    def mask_opd(self):
        """Mask from generated OPD map"""
        mask = np.ones(self.opd.shape)
        mask[self.opd==0] = 0
        return mask

    @property
    def coeff_pupil(self):
        """Pupil Zernike coefficients"""
        return self._coeff_pupil

    @property
    def coeff_segs(self):
        """Hexike coefficients for each segment"""
        return self._coeff_segs

    def mask_seg(self, i):
        """Return a sampled subsection of the analytic segment mask
        
        Parameters
        ----------
        i : int
            Segment index
        """
        # Mask off everything but the segment of interest
        pmask = self._mask_segs[i]
        outmask, _ = self._sample_mask(pmask)
        # Multiply by larger OPD pupil mask
        outmask *= self.mask_opd

        return self.opd_seg(i,outmask)

    def opd_seg(self, i, opd_pupil=None):
        """Return a subsection of some OPD image for the provided segment index
        
        Parameters
        ----------
        i : int
            Segment index
        opd_pupil : mask
            Optional pupil mask to modify resulting segment OPD
        """

        if opd_pupil is None: opd_pupil = self.opd_diff_pupil

        # Subsection the image, then pad to equal xy size
        x1,x2,y1,y2 = self.segment_positions[i]
        imsub = opd_pupil[y1:y2,x1:x2]

        (ny,nx) = imsub.shape
        npix_max = np.max(imsub.shape)

        # Make x and y the same size
        #imsub = pad_to_size(imsub, [npix_max,npix_max])
        imsub = pad_or_cut_to_size(imsub, [npix_max,npix_max])

        # If nx was larger than ny, then center y
        if nx > ny: # xhift in y
            shift = (ny-nx)//2 if y1>0 else (nx-ny)//2
            imsub = np.roll(imsub, shift, axis=0)
        elif ny > nx: # shift in x
            shift = (nx-ny)//2 if x1>0 else (ny-nx)//2
            imsub = np.roll(imsub, shift, axis=1)

        return imsub

    def _get_coeff_pupil(self):
        """Calculate Zernike coefficients for full OPD"""
        return self._get_coeff(self.opd, self.mask_opd, self.basis_zernike, iterations=2)

    def _get_opd_pupil(self):
        """Generate OPD for entire pupil based on coefficients"""
        opd = self._opd_from_coeff(self.coeff_pupil, self.basis_zernike, self.mask_pupil)
        return opd * self.mask_opd

    def _get_coeff_segs(self, opd_pupil=None, nterms=30):
        """Calculate Hexike coeffiencts each individual segment"""
        coeff_list = []
        for i in range(self.nseg): #, pmask in enumerate(self._mask_segs):
            mask_sub = self.mask_seg(i)
            opd_sub  = self.opd_seg(i, opd_pupil)
    
            coeff = self._get_coeff(opd_sub, mask_sub, self.basis_hexike, nterms=nterms)
            coeff_list.append(coeff)
        return coeff_list

    def _get_opd_new_segs(self, coeff_segs=None):
        """Generate segment OPDs in a list"""
        if coeff_segs is None: coeff_segs = self.coeff_segs

        opd_list = []
        for i, pmask in enumerate(self._mask_segs):
            mask_sub = self.mask_seg(i)

            opd = self._opd_from_coeff(coeff_segs[i], self.basis_hexike, mask_sub)
            opd *= mask_sub
            opd_list.append(opd)
    
        return opd_list

    def combine_opd_segs(self, opd_segs=None):
        """Combine list of OPD segments into a single image"""
        if opd_segs is None: opd_segs = self.opd_new_segs

        opd = np.zeros(self.opd.shape)
        for i, opd_sub in enumerate(opd_segs):
            pmask = self._mask_segs[i]
            mask, _ = self._sample_mask(pmask)
            mask *= self.mask_opd

            opd[mask==1] = opd_sub[opd_sub!=0]
    
        return opd

    def _get_coeff(self, opd, mask, basis, **kwargs):
        return zernike.opd_expand_nonorthonormal(opd, mask, basis=basis, **kwargs)

    def _opd_from_coeff(self, coeff, basis, mask):
        """Generate OPD image from a set of coefficients, basis function, and mask"""
        npix = mask.shape[0]
        return zernike.opd_from_zernikes(coeff, basis=basis, npix=npix, outside=0)
    
    def _sample_mask(self, mask):
        """Sample an analytic pupil mask at 1024x1024"""
        npix = self.npix
        outmask, pixelscale = mask.sample(npix=npix-8, return_scale=True)
        #outmask = pad_to_size(outmask, [npix,npix])
        outmask = pad_or_cut_to_size(outmask, [npix,npix])

        return outmask, pixelscale

    def _get_seg_xy(self, i):
        """
        Get the xy pixel indices (range) of a particular segment mask.
        Returns (x1,x2,y1,y2)
        """
        pmask = self._mask_segs[i]
        outmask, pixelscale = self._sample_mask(pmask)

        pix_rad = np.int(np.ceil(((pmask.side+pmask.gap) / pixelscale).value))
        pix_cen = pmask._hex_center(i+1)
        xc = pix_cen[1] / pixelscale.value + outmask.shape[1] / 2
        yc = pix_cen[0] / pixelscale.value + outmask.shape[0] / 2

        # Grab the pixel ranges
        x1 = np.int(xc - pix_rad)
        x2 = x1 + 2*pix_rad
        y1 = np.int(yc - pix_rad)
        y2 = y1 + 2*pix_rad

        # Limits on x/y positions
        x1 = np.max([x1,0])
        x2 = np.min([x2,self.npix])
        y1 = np.max([y1,0])
        y2 = np.min([y2,self.npix])

        return (x1,x2,y1,y2)

def opd_extract_helper(args):
    return OPD_extract(args[0], args[1], verbose=False)

def opd_extract_mp(opds, header, nproc=None):
    """Generate multiple OPD_extract objects
    
    If you have multiple opd maps to perform Zernike/Hexike
    extractions on, this function calls Python's multiprocessor
    routines to perform calculations in parallel.
    
    Parameters
    ----------
    opds : ndarray
        Data cube of size (nopds, ny, nx)
    header : obj
        Header information
    nproc : int, None
        Number of simultaneous processes to perform.
        If not set, then: 
        
        >>> nproc = int(np.min([nopd,mp.cpu_count()*0.75]))
    """
    if nproc is None:
        nopd = opds.shape[0]
        nproc = int(np.min([nopd,mp.cpu_count()*0.75]))

    pool = mp.Pool(nproc)
    worker_arguments = [(opds[i,:,:],header) for i in range(nopd)]
    opds_all = pool.map(opd_extract_helper, worker_arguments)
    pool.close()
    pool.join()

    return opds_all

def opd_sci_gen(opd):
    """OPD image for science observation
    
    Function to go through an OPD class and generate a science OPD image.
    
    Parameters
    ----------
    opd : obj
        OPD_extract object    
    """
    mask_pupil = opd.mask_pupil * opd.mask_opd
    opd_pupil  = opd.opd_new_pupil * mask_pupil
    opd_segs   = opd.combine_opd_segs(opd.opd_new_segs) * mask_pupil
    opd_resid  = (opd.opd - (opd_pupil+opd_segs)) * mask_pupil

    # OPD for science target observation
    opd_sci = opd_pupil + opd_segs + opd_resid

    # Return the residuals for use in drifted
    return (opd_sci, opd_resid)

def opd_sci_gen_mp(opds_all, nproc=None):
    """Multiple OPD images for science observations
    
    Function to go through multiple OPD classes and generate
    science OPD images.
    
    Parameters
    ----------
    opds_all : obj list
        List of OPD_extract objects
    nproc : int, None
        Number of simultaneous processes to perform.
        If not set, then: 
        
        >>> nproc = int(np.min([nopd,mp.cpu_count()*0.75]))
    """

    if nproc is None:
        nopd = len(opds_all)
        nproc = int(np.min([nopd,mp.cpu_count()*0.75]))

    pool = mp.Pool(nproc)
    output = pool.map(opd_sci_gen, opds_all)
    pool.close()
    pool.join()

    #opd_sci_list, opd_resid_list = zip(*output)
    return zip(*output)

def opd_ref_gen(args, verbose=False):
    """Generate a drifted OPD image"""
    opd, wfe_drift, pup_cf_std, seg_cf_std, opd_resid, case = args
    np.random.seed()

    # Pupil mask (incl. segment spaces and secondary struts)
    mask_pupil = opd.mask_pupil * opd.mask_opd

    # Case of Pupil Only
    if case==1:
        var = pup_cf_std**2
        std = np.sqrt(var/var.sum()) * (wfe_drift/1000)
        rand = np.random.randn(std.size) * std
        coeff_pupil_new = opd.coeff_pupil + rand #delta_coeff[:pup_var.size]
        ncf = pup_cf_std[pup_cf_std!=0].size

        opd_pupil_new = opd._opd_from_coeff(coeff_pupil_new, opd.basis_zernike, mask_pupil) * mask_pupil
        opd_segs_new  = opd.combine_opd_segs(opd.opd_new_segs) * mask_pupil

        # Total
        opd_sci = opd.opd * mask_pupil
        opd_ref = opd_pupil_new + opd_segs_new + opd_resid

        # Take the difference and make sure the RMS WFE difference is correct
        # Add this to the overall pupil coefficients
        opd_diff = (opd_sci - opd_ref) * mask_pupil
        rms_diff = opd_diff[opd_diff!=0].std()
        delta_rms = wfe_drift/1000 - rms_diff
        ind = var>0
        coeff_pupil_new[ind] += 1.1 * delta_rms * (rand[ind]/np.abs(rand[ind])) / np.sqrt(var[ind].size)
        opd_pupil_new = opd._opd_from_coeff(coeff_pupil_new, opd.basis_zernike, mask_pupil) * mask_pupil

        opd_ref = opd_pupil_new + opd_segs_new + opd_resid

    # Case of Segments Only
    elif case==2:
        opd_pupil_new = opd.opd_new_pupil # No drift to overall pupil

        # Segments
        # Random Gaussian noise distributed across each Zernike coeff
        rand_all = []
        coeff_segs_new = []
        var = seg_cf_std**2
        std = np.sqrt(var/var.sum()) * (wfe_drift/1000)
        for cf in opd.coeff_segs:
            rand = np.random.randn(std.size) * std
            rand_all.append(rand)
            coeff_segs_new.append(cf + rand)

        opd_segs_new_list = opd._get_opd_new_segs(coeff_segs_new)
        opd_segs_new = opd.combine_opd_segs(opd_segs_new_list) * mask_pupil

        # Total
        opd_sci = opd.opd * mask_pupil
        opd_ref = opd_pupil_new + opd_segs_new + opd_resid

        # Take the difference and make sure the RMS WFE difference is correct
        opd_diff = (opd_sci - opd_ref) * mask_pupil
        # Add this to the overall segment coefficients
        rms_diff = opd_diff[opd_diff!=0].std()
        delta_rms = wfe_drift/1000 - rms_diff
        ind = var>0; nind = var[ind].size
        for i,cf in enumerate(coeff_segs_new):
            rand = rand_all[i]
            cf[ind] += delta_rms * (rand[ind]/np.abs(rand[ind])) / np.sqrt(nind)
        opd_segs_new_list = opd._get_opd_new_segs(coeff_segs_new)
        opd_segs_new = opd.combine_opd_segs(opd_segs_new_list) * mask_pupil

        opd_ref = opd_pupil_new + opd_segs_new + opd_resid

    # Case of Pupil and Segments distributed evenly
    elif case==3:
        # Pupil
        var = pup_cf_std**2
        std = np.sqrt(var/var.sum()) * (wfe_drift/1000)
        rand_pup = np.random.randn(std.size) * std / np.sqrt(2.0)
        coeff_pupil_new = opd.coeff_pupil + rand_pup #delta_coeff[:pup_var.size]
        opd_pupil_new = opd._opd_from_coeff(coeff_pupil_new, opd.basis_zernike, mask_pupil) * mask_pupil

        # Segments
        # Random Gaussian noise distributed across each Zernike coeff
        coeff_segs_new = []
        var = seg_cf_std**2
        std = np.sqrt(var/var.sum()) * (wfe_drift/1000) / np.sqrt(2.0)
        for cf in opd.coeff_segs:
            rand = np.random.randn(std.size) * std
            coeff_segs_new.append(cf + rand)

        #coeff_segs_new = [cf + np.random.normal(0,wfe_drift/1000,cf.shape) for cf in opd.coeff_segs]
        opd_segs_new_list = opd._get_opd_new_segs(coeff_segs_new)
        opd_segs_new = opd.combine_opd_segs(opd_segs_new_list) * mask_pupil


        # Total
        opd_sci = opd.opd * mask_pupil
        opd_ref = opd_pupil_new + opd_segs_new + opd_resid

        # Take the difference and make sure the RMS WFE difference is correct
        # Add this to the overall pupil coefficients
        opd_diff = (opd_sci - opd_ref) * mask_pupil
        rms_diff = opd_diff[opd_diff!=0].std()
        delta_rms = wfe_drift/1000 - rms_diff
        ind = (pup_cf_std**2)>0; nind = rand_pup[ind].size
        coeff_pupil_new[ind] += 1.1 * delta_rms * (rand_pup[ind]/np.abs(rand_pup[ind])) / np.sqrt(nind)
        #coeff_pupil_new += delta_rms
        #coeff_pupil_new[pup_cf_std!=0] -= delta_rms
        opd_pupil_new = opd._opd_from_coeff(coeff_pupil_new, opd.basis_zernike, mask_pupil) * mask_pupil

        opd_ref = opd_pupil_new + opd_segs_new + opd_resid

    # Jeremy's Method
    elif case==4:
        opd_sci = opd.opd * mask_pupil
        opd_ref = opd_drift_nogood(opd_sci, wfe_drift) * mask_pupil


    if verbose:
        opd_sci = opd.opd * mask_pupil
        opd_diff = (opd_sci - opd_ref) * mask_pupil
        print('Sci RMS: {:.3f}, Ref RMS: {:.3f}, RMS diff: {:.4f}' \
              .format(opd_sci[opd_sci!=0].std(), opd_ref[opd_ref!=0].std(), opd_diff[opd_diff!=0].std()))

    return opd_ref

# Function to drift a list of  OPDs 
def ODP_drift_all(wfe_drift, opds_all, pup_cf_std, seg_cf_std, opd_resid_list):
    """Drift list of OPDs
    
    Drift a list of OPDs by some RMS WFE (multiprocess function).

    Parameters
    ----------
    wfe_drift : float
        Single value in nm
    opds_all : list
        List of OPD objects (usually 10 from RevV OPDs)
    pup_cf_std : float
        Zernike sigma for overall pupil
    seg_cf_std : float
        Hexike sigma for segments
    opd_resid_list : list
        List of residual images (from opd_sci_gen) for each OPD
    """

    case = 1 # Only drift full pupil; no segment drifts

    nopd = len(opds_all)
    nproc = int(np.min([nopd,mp.cpu_count()*0.75]))

    pool = mp.Pool(nproc)
    worker_arguments = [(opds_all[i], wfe_drift, pup_cf_std, seg_cf_std, 
                         opd_resid_list[i], case) for i in range(nopd)]
    out = pool.map(opd_ref_gen, worker_arguments)
    pool.close()
    pool.join()

    print('Finished: {:.0f} nm'.format(wfe_drift))

    return out


def get_psf(opd, header, filter='F410M', mask=None, pupil=None, 
    jitter=False, njit=10, jitter_sigma=0.005, r=None, theta=None, **kwargs):
    """Generate PSF
    
    Create a NIRCam PSF based on the input opd, header, and other info.

    The jitter model recomputes the PSF at random positional offsets with
    a standard deviation equal to jitter_sigma. It does this njit times
    and averages the resulting PSFs. 
    """

    hdu = fits.PrimaryHDU(opd)
    hdu.header = header.copy()
    opd_hdulist = fits.HDUList([hdu])    

    nc = webbpsf.NIRCam()
    nc.pupilopd   = opd_hdulist 
    nc.filter     = filter
    nc.image_mask = mask
    nc.pupil_mask = pupil
    nc.options['parity'] = 'odd'

    # Reposition stellar spot to optimal position on bar if LWB or SWB
    # If no bar mask, then this function returns 0s
    r_nom, theta_nom = offset_bar(filter,mask)
    r     = r_nom     if r     is None else r
    theta = theta_nom if theta is None else theta
    nc.options['source_offset_r']     = r
    nc.options['source_offset_theta'] = theta	
    #nc.options['output_mode'] = 'both'

    nc.options['jitter'] = None
    if jitter and (jitter_sigma>0):
        jit_type = 'Full'
        offset_x, offset_y = nrc_utils.rtheta_to_xy(r, theta)
        for i in range(njit):
            # Random position offset
            jit_off = np.random.normal(0, jitter_sigma)
            jit_ang = np.random.uniform(0, 180)
        
            # Convert jitter offset to (x,y) position and add to source location 
            jit_x, jit_y = nrc_utils.rtheta_to_xy(jit_off, jit_ang)
            off_x_new = offset_x + jit_x
            off_y_new = offset_y + jit_y
            rtemp, ttemp = nrc_utils.xy_to_rtheta(off_x_new, off_y_new)
            
            # Save position offsets
            nc.options['source_offset_r']     = rtemp
            nc.options['source_offset_theta'] = ttemp
            hdulist = nc.calc_psf(fov_arcsec=10)				
            if i==0:
                data0 = np.zeros_like(hdulist[0].data)
                data1 = np.zeros_like(hdulist[1].data)
            # Sum up the PSFs
            data0 += hdulist[0].data; data1 += hdulist[1].data
        
        # Average jittered data
        hdulist[0].data = data0 / njit
        hdulist[1].data = data1 / njit
    else:
#         fov_pix = 159
#         oversample = 4
#         im, im_over = nrc_utils.gen_image_coeff(filter, pupil=pupil, mask=mask,
#             fov_pix=fov_pix, oversample=oversample, return_oversample=True,
#             offset_r=r, offset_theta=theta, save=False, force=False, opd=opd_hdulist)
# 
#         hdu = fits.PrimaryHDU(im_over)
#         hdu.header['EXTNAME'] = ('OVERSAMP')
#         hdu.header['OVERSAMP'] = oversample
#         hdu.header['DET_SAMP'] = oversample
#         hdu.header['PIXELSCL'] = 0.063 / hdu.header['OVERSAMP']
#         hdulist = fits.HDUList([hdu])
#         
#         hdu = fits.PrimaryHDU(im)
#         hdu.header['EXTNAME'] = ('DET_SAMP')
#         hdu.header['OVERSAMP'] = 1
#         hdu.header['DET_SAMP'] = oversample
#         hdu.header['PIXELSCL'] = 0.063
#         hdulist.append(hdu)
        
        hdulist = nc.calc_psf(fov_arcsec=10)
        jit_type = 'None'
        jitter_sigma = 0

#     if jitter is not None:
#         if jitter.lower() == 'gaussian':
#             nc.options['jitter'] = 'gaussian'
#             nc.options['jitter_sigma'] = jitter_sigma
#             hdulist = nc.calc_psf(fov_arcsec=10)
#         elif jitter.lower() == 'full':
#             nc.options['jitter'] = None
#             njit = 10
#             for i in range(njit):
#                 rtemp = r + np.random.normal(0, jitter_sigma)
#                 ttemp = theta + np.random.uniform(0,360)
#                 nc.options['source_offset_r']     = rtemp
#                 nc.options['source_offset_theta'] = ttemp
#                 hdulist = nc.calc_psf(fov_arcsec=10)				
#                 if i==0:
#                     data0 = np.zeros_like(hdulist[0].data)
#                     data1 = np.zeros_like(hdulist[1].data)
#                 # Sum up the PSFs
#                 data0 += hdulist[0].data; data1 += hdulist[1].data
#             # Average jittered data
#             hdulist[0].data = data0 / njit
#             hdulist[1].data = data1 / njit
#         else:
#             raise ValueError('Do not recognize jitter = '+jitter)
#     else: # Create nominal PSF, no jitter
#         hdulist = nc.calc_psf(fov_arcsec=10)

    xval, yval = nrc_utils.rtheta_to_xy(r, theta)
    xval = 0 if np.isclose(0, xval, atol=1e-4) else xval
    yval = 0 if np.isclose(0, yval, atol=1e-4) else yval
    for hdr in [hdulist[0].header, hdulist[1].header]:
        hdr.insert('extname', ('offset',     r, 'Source offset from center in arcsec'))
        hdr.insert('extname', ('theta',  theta, 'Degrees CCW from inst +Y in sci frame'))
        hdr.insert('extname', ('xoffset', xval, 'Source xoffset in sci frame (arcsec)'))
        hdr.insert('extname', ('yoffset', yval, 'Source yoffset in sci frame (arcsec)'))
        hdr.insert('extname', ('JITRTYPE', jit_type, 'Type of jitter applied'))
        hdr.insert('extname', ('JITRSIGM', jitter_sigma, 'Gaussian sigma for jitter [arcsec]'))

    return hdulist


#def offset_bar(filt, mask):
#    """
#    
#    Get the appropriate offset in the x-position to place a source on a bar mask.
#    Each bar is 20" long with edges and centers corresponding to:
#      SWB: [1.03, 2.10, 3.10] (um) => [-10, 0, +10] (asec)
#      LWB: [2.30, 4.60, 6.90] (um) => [+10, 0, -10] (asec)
#    """
#
#    if (mask is not None) and ('WB' in mask):		
#        # What is the effective wavelength of the filter?	
#        #bp = pynrc.read_filter(filter)
#        #w0 = bp.avgwave() / 1e4
#        w0 = np.float(filt[1:-1])/100
#    
#        # Choose wavelength from dictionary
#        wdict = {'F182M': 1.84, 'F187N': 1.88, 'F210M': 2.09, 'F212N': 2.12, 
#                 'F250M': 2.50, 'F300M': 2.99, 'F335M': 3.35, 'F360M': 3.62, 
#                 'F410M': 4.09, 'F430M': 4.28, 'F460M': 4.63, 'F480M': 4.79,
#                 'F200W': 2.23, 'F277W': 3.14, 'F356W': 3.97, 'F444W': 4.99}
#        w = wdict.get(filt, w0)
#
#        # Get appropriate x-offset
#        #xoff_asec = np.interp(w,wpos,xpos)
#    
#        if 'SWB' in mask:
#            if filt[-1]=="W": xoff_asec = 6.83 * (w - 2.196)
#            else:             xoff_asec = 7.14 * (w - 2.100)
#        elif 'LWB' in mask:
#            if filt[-1]=="W": xoff_asec = -3.16 * (w - 4.747)
#            else:             xoff_asec = -3.26 * (w - 4.600)
#        
#        #print(w, xoff_asec)
#        
#        yoff_asec = 0.0
#    
#        r, theta = nrc_utils.xy_to_rtheta(xoff_asec, yoff_asec)
#    else:
#        r, theta = (0.0, 0.0)
#    
#    #print(r, theta)
#    return r, theta


def gen_psf_ref_all(opd_ref_list_all, opd_header, filt, mask, pupil, verbose=False, **kwargs):
    """Calculate a series of reference PSFs
    
    Given a list of reference OPD, calculate a series of PSFs.
    Assume that opd_ref_list_all is a nested list of drifted OPDs.
    For instance::
    
        [[10 OPDs with 1nm drift], [10 OPDs with 5nm drift], [10 OPDs with 10nm drift]]
    """
    psf_ref_all = []
    npsf = len(opd_ref_list_all)
    for i in range(npsf):
        opd_ref_list = opd_ref_list_all[i]
        psf_ref = [get_psf(opd, opd_header, filt, mask, pupil, **kwargs) for opd in opd_ref_list]
        psf_ref_all.append(psf_ref)
        if verbose: print('Finished {} of {}.'.format(i+1, npsf))

    return psf_ref_all

def get_contrast_old(psf0,psf1,psf2):
    """
    For science and reference PSFs, return the contrast curve.
    Assumes no noise other than residual speckle noise.
    psf0 is the planet PSF
    psf1 is the stellar PSF
    psf2 is the reference PSF
    """

    # PSF subtraction
    from copy import deepcopy
    psf_diff = deepcopy(psf1)
    psf_diff[0].data = (psf1[0].data - psf2[0].data)
    #psf_diff[1].data = (psf1[1].data - psf2[1].data)

    # Radial noise profiles of PSF difference
    #rr0, stds0 = webbpsf.radial_profile(psf_diff, ext=0, stddev=True)
    #rr1, stds1 = webbpsf.radial_profile(psf_diff, ext=1, stddev=True)

    ## Total planet signal at a radius of 0.5"
    #rr_psf0, mn_psf0, ee_psf0 = webbpsf.radial_profile(psf0, ext=0, EE=True)
    #rad_asec = 0.5
    #npix = np.pi * (rad_asec / psf0[0].header['PIXELSCL'])**2
    ## Get the encircled energy of planet at radius
    #planet_signal = np.interp(rad_asec, rr_psf0, ee_psf0)
    ## nsigma of planet signal relative to noise
    #contrast =  np.sqrt(stds0**2 * npix) / planet_signal

    data = psf_diff[0].data
    header = psf_diff[0].header
    pixelscale = header['PIXELSCL']
    
    rho = dist_image(data, pixscale=pixelscale)
    binsize = pixelscale
    bins = np.arange(rho.min(), rho.max() + binsize, binsize)
    rr0 = bins[:-1] + binsize / 2

    stds0 = binned_statistic(rho, data, func=np.std, bins=bins)
    contrast = stds0 / np.max(psf0[0].data)

    return rr0, contrast


def get_contrast(speckle_noise_image, planet_psf):
    """Output single contrast curve
    
    Return the contrast for curve for a speckle noise map.
    Both inputs are HDULists.
    """

    # Radial profile of the noise image
    ext = 0
    data = speckle_noise_image[ext].data
    header = speckle_noise_image[ext].header

    pixelscale = header['PIXELSCL']
    xoff = header['XOFFSET'] / pixelscale
    yoff = header['YOFFSET'] / pixelscale

    xcen = header['NAXIS1'] / 2.0 + xoff
    ycen = header['NAXIS2'] / 2.0 + yoff

    #rr0, stds0 = webbpsf.radial_profile(speckle_noise_image, ext=ext, center=(xcen,ycen))
    rho = dist_image(data, pixscale=pixelscale, center=(xcen,ycen))
    binsize = pixelscale
    bins = np.arange(rho.min(), rho.max() + binsize, binsize)
    rr0 = bins[:-1] + binsize / 2

    stds0 = binned_statistic(rho, data, func=np.mean, bins=bins)
    contrast = stds0 / np.max(planet_psf[0].data)

    #rr1, stds1 = webbpsf.radial_profile(psf_diff, ext=1)


    return rr0, contrast


def get_contrast_all_old(psf_planet, psf_star_all, psf_ref_all):
    """Multiple contrast curves
    
    Return a list of contrast curves based on a list of PSFs
    """
    nopd = len(psf_star_all)
    return [get_contrast_old(psf_planet,psf_star_all[i],psf_ref_all[i]) for i in range(nopd)]

def residual_speckle_image(psf_star_all, psf_ref_all):
    """Create residual speckle images
    
    For a list of science and reference PSFs, create residual speckle images
    """
    nopd = len(psf_star_all)
    psf1 = psf_star_all[0] # Stellar PSF
    psf2 = psf_ref_all[0]  # Reference PSF

    # Do the difference for each extension (pixel-sampled and oversampled)
    ny0,nx0 = psf1[0].data.shape
    ny1,nx1 = psf1[1].data.shape

    diff_all0 = np.zeros([nopd,ny0,nx0])
    diff_all1 = np.zeros([nopd,ny1,nx1])

    for i in range(nopd):
        diff_all0[i,:,:] = psf_star_all[i][0].data - psf_ref_all[i][0].data
        diff_all1[i,:,:] = psf_star_all[i][1].data - psf_ref_all[i][1].data

    return diff_all0, diff_all1

def speckle_noise_image(psf_star_all, psf_ref_all):
    """Create a speckle noise image map
    
    For a list of science and reference PSFs, create a speckle noise image map
    """

    nopd = len(psf_star_all)
    psf1 = psf_star_all[0] # Stellar PSF
    psf2 = psf_ref_all[0]  # Reference PSF

    # Do the difference for each extension (pixel-sampled and oversampled)
    ny0,nx0 = psf1[0].data.shape
    ny1,nx1 = psf1[1].data.shape

    diff_all0 = np.zeros([nopd,ny0,nx0])
    diff_all1 = np.zeros([nopd,ny1,nx1])

    for i in range(nopd):
        diff_all0[i,:,:] = psf_star_all[i][0].data - psf_ref_all[i][0].data
        diff_all1[i,:,:] = psf_star_all[i][1].data - psf_ref_all[i][1].data

    # Create HDU to hold standard deviations
    from copy import deepcopy
    psf_diff = deepcopy(psf1)

    # Save the standard deviation of the speckle noise for each extension
    psf_diff[0].data = diff_all0.std(axis=0)
    psf_diff[1].data = diff_all1.std(axis=0)

    return psf_diff


def opd_drift_nogood(opd, drift, nterms=8, defocus_frac=0.8):
    """
    Add some WFE drift (in nm) to an OPD image.

    Parameters
    ------------
    opd : ndarray
        OPD images (can be an array of images).
    header : obj
        Header file
    drift : float
        WFE drift in nm

    Returns
    --------
    HDUList :
        Returns an HDUList, which can be passed to webbpsf
    """


    # Various header info for coordinate grid
    diam  = 6.55118110236
    pix_m = 0.00639763779528
    wfe_rm = 132.

    # Create x/y carteesian grid
    sh = opd.shape
    if len(sh) == 3: 
        nz,ny,nx = sh
        opd_sum = opd.sum(axis=0)
        mask0 = (opd_sum == 0)
        mask1 = (opd_sum != 0)
    else:
        nz = 1
        ny,nx = sh
        # Masks for those values equal (and not) to 0
        mask0 = (opd == 0)
        mask1 = (opd != 0)

    y,x = np.indices((ny,nx))
    center = tuple((a - 1) / 2.0 for a in [nx,ny])
    y = y - center[1]; x = x - center[0]
    y *= pix_m; x *= pix_m

    # Convert to polar coordinates
    rho = np.sqrt(x**2 + y**2) / (diam / 2)
    theta = np.arctan2(y,x)

    # Generate Zernike maps
    # Drop the piston
    zall = (poppy.zernike.zernike_basis(nterms, rho=rho, theta=theta))[1:,:,:]
    zall[:,mask0] = 0

    # Sum Zernikes and normalize to total
    # Exclude defocus term
    zmost = np.concatenate((zall[0:2,:,:], zall[3:,:,:]))
    ztot = zmost.sum(axis=0)
    ztot /= ztot.sum()

    # Normalize defocus
    zfoc = zall[2,:,:]
    zfoc /= zfoc.sum()

    # Fraction of total that goes into defocus
    zsum = (1.0-defocus_frac)*ztot + defocus_frac*zfoc

    # Set masked pixels to 0 and normalize to unmasked sigma
    zsum[mask0] = 0
    zsum /= zsum[mask1].std()

    # RMS factor measured versus ideal
    # Accounts for unit differences as well (meters vs nm)
    # header['WFE_RMS'] is in nm, as is drift
    if len(sh) == 3:
        rms_opd = np.array([(opd[i,:,:])[mask1].std() for i in range(nz)])
        rms_fact = rms_opd / wfe_rm
        drift_act = rms_fact * drift
        zadd = zsum * drift_act.reshape([nz,1,1]) # Array broadcasting
    else:
        drift_act = drift * opd[mask1].std() / wfe_rm
        zadd = zsum * drift_act

    return opd + zadd

def _opd_drift_nogood(opd, header, drift, nterms=8, defocus_frac=0.8):
    """
    Add some WFE drift (in nm) to an OPD image.

    Parameters
    ------------
    opd : ndarray
        OPD images (can be an array of images).
    header : obj
        Header file
    drift : float
        WFE drift in nm

    Returns
    --------
    HDUList :
        Returns an HDUList, which can be passed to webbpsf
    """


    # Various header info for coordinate grid
    diam = header['PUPLDIAM']
    pix_m = header['PUPLSCAL']

    # Create x/y carteesian grid
    sh = opd.shape
    if len(sh) == 3: 
        nz,ny,nx = sh
        opd_sum = opd.sum(axis=0)
        mask0 = (opd_sum == 0)
        mask1 = (opd_sum != 0)
    else:
        nz = 1
        ny,nx = sh
        # Masks for those values equal (and not) to 0
        mask0 = (opd == 0)
        mask1 = (opd != 0)

    y,x = np.indices((ny,nx))
    center = tuple((a - 1) / 2.0 for a in [nx,ny])
    y = y - center[1]; x = x - center[0]
    y *= pix_m; x *= pix_m

    # Convert to polar coordinates
    rho = np.sqrt(x**2 + y**2) / diam
    theta = np.arctan2(y,x)

    # Generate Zernike maps
    zall = (poppy.zernike.zernike_basis(nterms, rho=rho, theta=theta))[1:,:,:]

    # Sum Zernikes and noramlize to total
    ztot = zall.sum(axis=0)
    ztot /= ztot.sum()

    # Fraction of total that goes into defocus
    ztot = (1.0-defocus_frac)*ztot + defocus_frac*(zall[2,:,:]/zall[2,:,:].sum())

    # Set masked pixels to 0 and normalize to unmasked sigma
    ztot[mask0] = 0
    ztot /= ztot[mask1].std()

    # RMS factor measured versus ideal
    # Accounts for unit differences as well (meters vs nm)
    # header['WFE_RMS'] is in nm, as is drift
    if len(sh) == 3:
        rms_opd = np.array([(opd[i,:,:])[mask1].std() for i in range(nz)])
        rms_fact = rms_opd / header['WFE_RMS']
        drift_act = rms_fact * drift
        zadd = ztot * drift_act.reshape([nz,1,1]) # Array broadcasting
    else:
        drift_act = drift * opd[mask1].std() / header['WFE_RMS']
        zadd = ztot * drift_act

    # Add Zernike drifts to original OPD
    hdu = fits.PrimaryHDU(opd + zadd)
    hdu.header = header.copy()
    hdu.header['history'] = 'Added {:.1f} nm of WFE drift.'.format(drift)
    hdulist = fits.HDUList([hdu])
    
    return hdulist

def read_opd_file(opd_file, opd_path=None, header=True):
    """
    Read in the OPD file data and header (optional)
    """
    
    if opd_path is None:
        data_path = webbpsf.utils.get_webbpsf_data_path() + '/'
        opd_path = data_path + 'NIRCam/OPD/'
        
    return fits.getdata(opd_path + opd_file, header=header)
        
def read_opd_slice(pupilopd, opd_path=None, header=True):
    """
    Get only a specified OPD slice from file.
    
    pupilopd is a tuple of form (filename, slice)
    """
    
    opd_file, slice = pupilopd
    out = read_opd_file(opd_file, opd_path, header)
    
    if header:
        data, header = out
        return data[slice,], header
    else:
        return out[slice,]


from __future__ import division, print_function, unicode_literals

from astropy.convolution import convolve, convolve_fft
from astropy.convolution import Gaussian2DKernel, Gaussian1DKernel
from scipy.ndimage.interpolation import rotate
from scipy import fftpack
from copy import deepcopy

# Import libraries
from . import *
from .nrc_utils import *

import logging
_log = logging.getLogger('pynrc')

eps = np.finfo(float).eps

class obs_hci(NIRCam):
    """NIRCam coronagraphy (and direct imaging)
    
    Subclass of the NIRCam instrument class used to observe stars 
    (plus exoplanets and disks) with either a coronagraph or direct 
    imaging. 
    
    The main concept is to generate a science target of the primary
    source along with a simulated disk structure. Planets are further
    added to the astronomical scene. A separate reference source is 
    also defined for PSF subtraction, which contains a specified WFE. 
    A variety of methods exist to generate slope images and analyze 
    the PSF-subtracted results via images and contrast curves.

    Parameters
    ----------
    sp_sci : :mod:`pysynphot.spectrum`
        A pysynphot spectrum of science target (e.g., central star).
        Should already be normalized to the apparent flux.
    sp_ref : :mod:`pysynphot.spectrum`
        A pysynphot spectrum of reference target.
        Should already be normalized to the apparent flux.
    distance : float
        Distance in parsecs to the science target. This is used for
        flux normalization of the planets and disk.
    wfe_ref_drift: float
        WFE drift in nm between the science and reference targets.
        Expected values are between ~3-10 nm.
    wfe_ref_drift: float
        WFE drift in nm between science roll angles. Default=0.
    wind_mode : str
        'FULL', 'STRIPE', or 'WINDOW'
    xpix : int
        Size of the detector readout along the x-axis. The detector is
        assumed to be in window mode  unless the user explicitly 
        sets wind_mode='FULL'.
    ypix : int
        Size of the detector readout along the y-axis. The detector is
        assumed to be in window mode  unless the user explicitly 
        sets wind_mode='FULL'.
    disk_hdu : HDUList
        A model of the disk in photons/sec. This requires header
        keywords PIXSCALE (in arcsec/pixel) and DISTANCE (in pc).
        
    """
    
    def __init__(self, sp_sci, sp_ref, distance, wfe_ref_drift=10, wfe_roll_drift=0,
        offset_list=None, wind_mode='WINDOW', xpix=320, ypix=320, disk_hdu=None, 
        verbose=False, **kwargs):
                 
        if 'FULL'   in wind_mode: xpix = ypix = 2048
        if 'STRIPE' in wind_mode: xpix = 2048
                
        #super(NIRCam,self).__init__(**kwargs)
        # Not sure if this works for both Python 2 and 3
        NIRCam.__init__(self, wind_mode=wind_mode, xpix=xpix, ypix=ypix, **kwargs)

        if (wind_mode=='FULL') and (self.channel=='SW'):
            raise NotImplementedError('SW Full Frame not yet implemented.')

        # Spectral models
        self.sp_sci = sp_sci
        self.sp_ref = sp_ref
        self._wfe_ref_drift = wfe_ref_drift
        self.wfe_roll_drift = wfe_roll_drift

        
        # Distance to source in pc
        self.distance = distance
        self._planets = []

        # Background/off-axis PSF coeff updates
        # -------------------------------------
        # Background PSF should be same size as primary PSF
        # Used to generate position-dependent off-axis PSFs
        if self.mask is not None:
            if verbose: print("Generating background PSF coefficients...")
            log_prev = conf.logging_level
            self._fov_pix_bg = self.psf_info['fov_pix']
            setup_logging('WARN', verbose=False)
            self.update_psf_coeff()
            setup_logging(log_prev, verbose=False)
            
        # Create mask throughput images seen by each detector
        self._gen_cmask()

        # Cached PSFs
        # -----------
        # Generate cached PSFs for quick retrieval.
        if verbose: print("Generating oversampled PSFs...")
        if self.mask is None:
            _, psf = self.gen_psf(return_oversample=True, use_bg_psf=False)
            self.psf_center_over = psf
            self.psf_offaxis_over = self.psf_center_over
        elif self.mask[-1]=='R':
            _, psf = self.gen_psf(return_oversample=True, use_bg_psf=False)
            self.psf_center_over = psf
            _, psf = self.gen_psf(return_oversample=True, use_bg_psf=True)
            self.psf_offaxis_over = psf
        elif self.mask[-1]=='B':
            # Bar mask "central" PSFs are a list of PSFs along the mask
            self._gen_psfbar_list()
            _, psf = self.gen_psf(return_oversample=True, use_bg_psf=True)
            self.psf_offaxis_over = psf
        
        # PSFs at each offset position
        # Only necessary if there is a disk or extended object.
        if disk_hdu is not None:
            if verbose: print("Generating PSFs for disk convolution...")

            # Offsets positions to build PSFs as we move away from mask
            # For bar masks, these PSFs are offset along the center of the bar.
            if self.mask is None:
                # if no coronagraphic mask, then only 1 PSF
                self.offset_list = [0.0]
                if offset_list is not None:
                    print('No coronagraph, so offset_list automatically set to [0.0].')
            elif offset_list is None:
                self.offset_list = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 1.0, 0.75, 1.5, 2.0, 5.0]
            else:
                self.offset_list = offset_list

            self._gen_psf_list()
        
        self._gen_ref(verbose=verbose)
        self._set_xypos()
        
        # Rescale input disk image to observation parameters
        self._disk_hdulist_input = disk_hdu
        self._gen_disk_hdulist()
        
        if verbose: print("Finished.")

    @property
    def wfe_drift(self):
        """WFE drift relative to nominal PSF (nm)"""
        return self._wfe_drift
    @wfe_drift.setter
    def wfe_drift(self, value):
        """Set the WFE drift value and update coefficients"""
        # Only update if the value changes
        _log.warning("Are you sure you don't mean wfe_ref_drift?")
        vold = self._wfe_drift; self._wfe_drift = value
        if vold != self._wfe_drift: 
            self.update_psf_coeff(wfe_drift=self._wfe_drift)
    
    def _set_wfe_drift(self, value, no_warn=True):
        """
        Similar to wfe_drift setter, but prevents warnings.
        Kind of a kludge. Make a better solution later?
        """
        if no_warn:
            log_prev = conf.logging_level
            setup_logging('ERROR', verbose=False)
            self.wfe_drift = value
            setup_logging(log_prev, verbose=False)
        else:
            self.wfe_drift = value

        
    @property
    def wfe_ref_drift(self):
        """WFE drift (nm) of ref obs relative to sci obs"""
        return self._wfe_ref_drift
    @wfe_ref_drift.setter
    def wfe_ref_drift(self, value):
        """Set the WFE drift value (updates self.nrc_ref)"""
        # Only update if the value changes
        vold = self._wfe_ref_drift; self._wfe_ref_drift = value
        if vold != self._wfe_ref_drift: 
            self._gen_ref()
            
    def gen_offset_psf(self, offset_r, offset_theta, sp=None, return_oversample=False):
        """Create a PSF offset from center FoV
        
        Generate some off-axis PSF at a given (r,theta) offset from center.
        This function is mainly for coronagraphic observations where the
        off-axis PSF varies w.r.t. position. The PSF is centered in the 
        resulting image.
        
        Parameters
        ----------
        offset_r : float
            Radial offset of the target from center in arcsec.
        offset_theta : float
            Position angle for that offset, in degrees CCW (+Y).
        
        Keyword Args
        ------------
        sp : None, :mod:`pysynphot.spectrum`
            If not specified, the default is flat in phot lam 
            (equal number of photons per spectral bin).
        return_oversample : bool
            Return either the pixel-sampled or oversampled PSF.

        """
        
        if sp is None:
            psf_center  = None #self.psf_center_over
            psf_offaxis = None #self.psf_offaxis_over
        else:
            if self.mask is None:
                # Direct imaging; both PSFs are the same
                _, psf_center = self.gen_psf(sp, return_oversample=True, use_bg_psf=False)
                psf_offaxis = psf_center
            elif self.mask[-1]=='R': # Round masks
                _, psf_center  = self.gen_psf(sp, return_oversample=True, use_bg_psf=False)
                _, psf_offaxis = self.gen_psf(sp, return_oversample=True, use_bg_psf=True)
            elif self.mask[-1]=='B': # Bar masks

                # Original offset value for observation
                baroff_orig = self.bar_offset
                # Specified offset along x-axis
                offx_asec, offy_asec = rtheta_to_xy(offset_r, offset_theta)
                self.bar_offset = offx_asec
                _, psf_center = self.gen_psf(sp, return_oversample=True, use_bg_psf=False)
                # Return to original bar offset position
                self.bar_offset = baroff_orig

                _, psf_offaxis = self.gen_psf(sp, return_oversample=True, use_bg_psf=True)

        psf = self._psf_lin_comb(offset_r, offset_theta, psf_center, psf_offaxis)
        if return_oversample:
            return psf
        else:
            fov_pix = self.psf_info['fov_pix']
            return krebin(psf, (fov_pix,fov_pix))

        
    def _psf_lin_comb(self, offset_r, offset_theta, psf_center=None, psf_offaxis=None):
        """
        Linearly combine off-axis and occulted PSFs.
        Returns an oversampled PSF.
        
        If passing the two PSFs, make sure they are the oversampled
        versions.
        
        Parameters
        ----------
        offset_r : float
            Radial offset of the target from center in arcsec.
        offset_theta : float
            Position angle for that offset, in degrees CCW (+Y).
            
        Keywords
        --------
        psf_center : None or ndarray
            Oversampled center (occulted) PSF. Defaults to flat spectral source.
        psf_offaxis : None or ndarray
            Oversampled off-axis (unocculted) PSF. Defaults to flat spectral source.
        """
        
        fov_pix        = self.psf_info['fov_pix']
        oversample     = self.psf_info['oversample']
        pixscale       = self.pix_scale
        fov_asec       = fov_pix * pixscale
        pixscale_over = pixscale / oversample
        
        # For coronagraphic masks, the offset PSF is well-determined by a linear
        # combination of the perfectly centered PSF and the off-axis PSF.
        if self.mask is None: # Direct imaging
            # Oversampled PSFs
            res  = self.psf_center_over  if psf_center  is None else psf_center
        elif self.mask[-1]=='R': # Round masks
            # Oversampled PSFs
            psf_center  = self.psf_center_over  if psf_center  is None else psf_center
            psf_offaxis = self.psf_offaxis_over if psf_offaxis is None else psf_offaxis

            # Oversampled image mask
            im_mask = coron_trans(self.mask, fov=fov_asec, pixscale=pixscale_over, nd_squares=False)
            im_mask = pad_or_cut_to_size(im_mask, psf_center.shape)

            ny, nx = im_mask.shape
            xv = (np.arange(nx) - nx/2) * pixscale_over
            a = np.interp(offset_r, xv, im_mask[ny//2,:]**2)
            b = 1 - a
            
            res = psf_offaxis*a + psf_center*b
        elif self.mask[-1]=='B': # Bar masks
            # Oversampled PSFs
            psf_offaxis = self.psf_offaxis_over if psf_offaxis is None else psf_offaxis
            ny, nx = psf_offaxis.shape

            # Determine x and y location
            offx_asec, offy_asec = rtheta_to_xy(offset_r, offset_theta)
            
            # If outside the 20" mask region, then just region the off-axis PSF
            if np.abs(offx_asec) > 10:
                return psf_offaxis

            # Get center PSF
            if psf_center is None:
                vals = self.psf_center_offsets
                arr = np.array(self.psf_center_over)
                func = interp1d(vals, arr, axis=0, kind='linear')
                psf_center = func(offx_asec)
            
            # Oversampled image mask
            im_mask = coron_trans(self.mask, fov=fov_asec, pixscale=pixscale_over, nd_squares=False)
            im_mask = pad_or_cut_to_size(im_mask, (ny,nx))
            xloc = int(offx_asec / pixscale_over + nx/2)
            mask_cut = im_mask[:,xloc]

            # Interpolate along the horizontal cut
            yv = (np.arange(ny) - ny/2) * pixscale_over
            a = np.interp(offy_asec, yv, mask_cut**2)
            b = 1 - a
            
            res = psf_offaxis*a + psf_center*b

        return res
        

    def _gen_disk_hdulist(self):
        """Create a correctly scaled disk model image.
        
        Also shifts relative to bar offset position.
        """
        if self._disk_hdulist_input is None:
            self.disk_hdulist = None
        else:
            xpix = self.det_info['xpix']
            ypix = self.det_info['ypix']
            oversample = self.psf_info['oversample']
        
            disk_hdul = self._disk_hdulist_input
            hdr = disk_hdul[0].header
            
            # Get rid of the central star flux
            # and anything interior to a few pixls
            image = disk_hdul[0].data
            image_rho = dist_image(image)
            image[image_rho < 4] = 0
                       
            # Resample disk to detector pixel scale
            # args_in  = (input pixelscale,  input distance)
            # args_out = (output pixelscale, output distance)
            args_in = (hdr['PIXELSCL'], hdr['DISTANCE'])
            args_out = (self.pixelscale, self.distance)            
            hdulist_out = image_rescale(disk_hdul, args_in, args_out, cen_star=False)
            
            # Expand to full observation size
            hdulist_out[0].data = pad_or_cut_to_size(hdulist_out[0].data, (ypix,xpix))
            self.disk_hdulist = hdulist_out
                        
        
    def _gen_ref(self, verbose=False):
        """Function to generate Reference observation class"""

        # PSF information
        opd = self.psf_info['opd']
        fov_pix = self.psf_info['fov_pix']
        oversample = self.psf_info['oversample']

        # Detector information
        wind_mode = self.det_info['wind_mode']
        xpix = self.det_info['xpix']
        ypix = self.det_info['ypix']
        
        offset_r = self.psf_info['offset_r']
        offset_theta = self.psf_info['offset_theta']
        
        # Create a NIRCam reference class
        # If it already exists, just update OPD info
        try:
            if verbose: print("Updating NIRCam reference coefficients...")
            self.nrc_ref.wfe_drift = self.wfe_ref_drift
        except AttributeError:
            if verbose: print("Creating NIRCam reference class...")
            nrc = NIRCam(self.filter, self.pupil, self.mask, module=self.module, \
                         wind_mode=wind_mode, xpix=xpix, ypix=ypix, \
                         fov_pix=fov_pix, oversample=oversample, opd=opd,
                         offset_r=offset_r, offset_theta=offset_theta,
                         wfe_drift=self.wfe_ref_drift, bar_offset=self.bar_offset)
            self.nrc_ref = nrc

        
    def _gen_psf_list(self):
        """
        Save instances of NIRCam PSFs that are incrementally offset 
        from coronagraph center to convolve with a disk image.
        """
        
        # If no mask, then the PSF looks the same at all radii
        self.psf_list = []
        if self.mask is None:
            psf = self.gen_offset_psf(0, 0)
            self.psf_list = [psf]
        elif self.mask[-1]=='R': # Round masks
            self.psf_list = [self.gen_offset_psf(offset, 0) for offset in self.offset_list]
        elif self.mask[-1]=='B': # Bar masks
            # Set bar offset to 0, for this part, then return to original value
            baroff_orig = self.bar_offset
            self.bar_offset = 0
            self.psf_list = [self.gen_offset_psf(offset, 0) for offset in self.offset_list]
            self.bar_offset = baroff_orig

    def _gen_psfbar_list(self):
        """
        Create instances of NIRCam PSFs that are incrementally offset 
        along the center of a coronagraphic wedge mask.
        """
        
        # Check that a bar mask is selected, otherwise exit
        if (self.mask is None) or (not self.mask[-1]=='B'):
            _log.warning('Bar mask not currently set (self.mask={}). Returning.'\
                         .format(self.mask))
            return
        
        xoff_arcsec_max = 10
        
        # Detector size
        xpix = self.det_info['xpix']
        xasec_half = np.ceil(xpix * self.pix_scale / 2)
        
        # Choose minimum of full field or selected window size
        xoff_asec = np.min([xasec_half, xoff_arcsec_max])
        
        # Offset values to create new PSF
        del_off = 2
        offset_vals = np.arange(-xoff_asec, xoff_asec+del_off, del_off)
        
        # Original offset value for observation
        baroff_orig = self.bar_offset
        
        # Loop through offset locations and save PSFs
        psf_list = []
        for offset in offset_vals:
            #print(offset)
            self.bar_offset = offset
            _, psf = self.gen_psf(return_oversample=True, use_bg_psf=False)
            psf_list.append(psf)
        
        # Return to original bar offset position
        self.bar_offset = baroff_orig
        
        self.psf_center_offsets = offset_vals
        self.psf_center_over = psf_list
            

    def _set_xypos(self):
        """
        Set x0 and y0 subarray positions.
        """
        
        wind_mode = self.det_info['wind_mode']
        if self.mask is not None:
            full = True if 'FULL' in wind_mode else False
            cdict = coron_ap_locs(self.module, self.channel, self.mask, full=full)
            xcen, ycen = cdict['cen']

            xpix, ypix = (self.det_info['xpix'], self.det_info['ypix'])
            if full: x0, y0 = (0,0)
            else: x0, y0 = (int(xcen-xpix/2), int(ycen-ypix/2))
        
            # Make sure subarray sizes don't push out of bounds
            if (y0 + ypix) > 2048: y0 = 2048 - ypix
            if (x0 + xpix) > 2048: x0 = 2048 - xpix
        
            self.update_detectors(x0=x0, y0=y0)
            
    def _gen_cmask(self, oversample=1):
        """
        Generate coronagraphic mask transmission images.
        
        Output images are in V2/V3 coordinates.
        """
        mask = self.mask
        module = self.module
        pixscale = self.pix_scale
        wind_mode = self.det_info['wind_mode']
        
        mask_dict = {}
        for det in self.Detectors:
            detid = det.detid
            
            if mask is None:
                mask_dict[detid] = None
            elif 'FULL' in wind_mode:
                mask_dict[detid] = build_mask_detid(detid, oversample, mask)
            else:
                fov = np.max([det.xpix, det.ypix]) * pixscale
                im = coron_trans(mask, module=module, pixscale=pixscale, 
                                 fov=fov, nd_squares=True)
                im = pad_or_cut_to_size(im, (det.ypix,det.xpix))
                mask_dict[detid] = im
            
        self.mask_images = mask_dict

    def planet_spec(self, Av=0, **kwargs):
        """Exoplanet spectrum.
        
        Return the planet spectrum from Spiegel & Burrows (2012) normalized
        to distance of current target. Output is a :mod:`pysynphot.spectrum`.
        
        Parameters
        ----------
        Av : float
            Extinction magnitude (assumes Rv=4.0).
            
        Keyword Args
        ------------
        atmo : str
            A string consisting of one of four atmosphere types:
            ['hy1s', 'hy3s', 'cf1s', 'cf3s'].
        mass: int
            Number 1 to 15 Jupiter masses.
        age: float
            Age in millions of years (1-1000).
        entropy: float
            Initial entropy (8.0-13.0) in increments of 0.25

        accr : bool
            Include accretion (default: False)?
        mmdot : float
            From Zhu et al. (2015), the Mjup^2/yr value.
            If set to None then calculated from age and mass.
        mdot : float
            Or use mdot (Mjup/yr) instead of mmdot.
        accr_rin : float
            Inner radius of accretion disk (units of RJup; default: 2)
        truncated: bool
             Full disk or truncated (ie., MRI; default: False)?
        """
        # Create planet class and convert to Pysynphot spectrum
        planet = planets_sb12(distance=self.distance, **kwargs)
        sp = planet.export_pysynphot()

        # Add extinction from the disk
        Rv = 4.0
        if Av>0: sp *= S.Extinction(Av/Rv,name='mwrv4')
        
        return sp
    

    @property
    def planets(self):
        """Planet info (if any exists)"""
        return self._planets

    def add_planet(self, atmo='hy3s', mass=10, age=100, entropy=10,
        xy=None, rtheta=None, runits='AU', Av=0, renorm_args=None, sptype=None,
        accr=False, mmdot=None, mdot=None, accr_rin=2, truncated=False):
        """Insert a planet into observation.
        
        Add exoplanet information that will be used to generate a point
        source image using a spectrum from Spiegel & Burrows (2012).
        Use self.kill_planets() to delete them.
        
        Coordinate convention is for +V3 up and +V2 to left.
        
        Parameters
        ----------
        atmo : str
            A string consisting of one of four atmosphere types:
            ['hy1s', 'hy3s', 'cf1s', 'cf3s'].
        mass: int
            Number 1 to 15 Jupiter masses.
        age: float
            Age in millions of years (1-1000).
        entropy: float
            Initial entropy (8.0-13.0) in increments of 0.25

        sptype : str
            Instead of using a exoplanet spectrum, specify a stellar type.        
        renorm_args : dict
            Pysynphot renormalization arguments in case you want
            very specific luminosity in some bandpass.
            Includes (value, units, bandpass). 
            
        Av : float
            Extinction magnitude (assumes Rv=4.0) of the exoplanet
            due to being embedded in a disk.
    
        xy : tuple, None
            (X,Y) position in V2/V3 coordinates of companion.
        rtheta : tuple, None
            Radius and position angle relative to stellar position. 
            Alternative to xy keyword
        runits : str
            What are the spatial units? Valid values are 'AU', 'asec', or 'pix'.

        accr : bool
            Include accretion (default: False)?
        mmdot : float
            From Zhu et al. (2015), the Mjup^2/yr value.
            If set to None then calculated from age and mass.
        mdot : float
            Or use mdot (Mjup/yr) instead of mmdot.
        accr_rin : float
            Inner radius of accretion disk (units of RJup; default: 2)
        truncated: bool
             Full disk or truncated (ie., MRI; default: False)?

         """
         
        if (xy is None) and (rtheta is None):
            _log.error('Either xy or rtheta must be specified.')

        if (xy is not None) and (rtheta is not None):
            _log.warning('Both xy and rtheta are specified. ' + \
                         'The xy keyword shall take priority.')

        # Size of subarray image in terms of pixels
        image_shape = (self.det_info['ypix'], self.det_info['xpix'])
        
        # XY location of planets within subarray with units from runits keyword
        loc = rtheta_to_xy(rtheta[0], rtheta[1]) if xy is None else xy

        # Define pixel location relative to the center of the subarray
        au_per_pixel = self.distance*self.pix_scale
        if 'AU' in runits:
            xoff, yoff = np.array(loc) / au_per_pixel
        elif ('asec' in runits) or ('arcsec' in runits):
            xoff, yoff = np.array(loc) / self.pix_scale
        elif ('pix' in runits):
            xoff, yoff = loc
        else:
            errstr = "Do not recognize runits='{}'".format(runits)
            raise ValueError(errstr)
            
        # Offset in terms of arcsec
        xoff_asec, yoff_asec = np.array([xoff, yoff]) * self.pix_scale
        _log.debug('(xoff,yoff) = {} pixels'.format((xoff,yoff)))
        _log.debug('(xoff_asec,yoff_asec) = {} arcsec'.format((xoff_asec,yoff_asec)))
    
        # Make sure planet is within image bounds
        sh_diff = np.abs(np.array([yoff,xoff])) - np.array(image_shape)/2
        if np.any(sh_diff>=0):
            _log.warning('xoff,yoff = {} is beyond image boundaries.'.format((xoff,yoff)))
            
        # X and Y pixel offsets from center of image
        # Dictionary of planet info
        if sptype is None:
            d = {'xyoff_pix':(xoff,yoff), 'atmo':atmo, 'mass':mass, 'age':age, 
                 'entropy':entropy, 'Av':Av, 'renorm_args':renorm_args, 
                 'accr':accr, 'mmdot':mmdot, 'mdot':mdot, 'accr_rin':accr_rin, 
                 'truncated':truncated}
        else:
            d = {'xyoff_pix':(xoff,yoff), 'sptype':sptype, 'Av':Av, 
                 'renorm_args':renorm_args}
        self._planets.append(d)
        
        
    def gen_planets_image(self, PA_offset=0, quick_PSF=True, use_cmask=False, **kwargs):
        """Create image of just planets.
        
        Use info stored in self.planets to create a noiseless slope image 
        of just the exoplanets (no star).
        
        Coordinate convention is for +V3 up and +V2 to left.
        
        Parameters
        ----------        
        PA_offset : float
            Rotate entire scene by some position angle.
            Positive values are counter-clockwise from +Y direction.
            Corresponds to instrument aperture PA.
        quick_PSF : bool
            Rather than generate a spectrum-weighted PSF, use the 
            cached PSF scaled by the photon count through the bandpass.
            Resulting PSFs are less accurate, but generation is much faster.
            Default is True.
        use_cmask : bool
            Use the coronagraphic mask image to determine if any planet is
            getting obscurred by a corongraphic mask feature
        """
        if len(self.planets)==0:
            _log.info("No planet info at self.planets")
            return 0.0
            
        if PA_offset is None: PA_offset=0
            
        image_shape = ypix, xpix = (self.det_info['ypix'], self.det_info['xpix'])
        image = np.zeros(image_shape)
        bar_offset = self.bar_offset
        bar_offpix = bar_offset / self.pixelscale
        for pl in self.planets:

            # Create slope image (postage stamp) of planet
            if pl.get('sptype') is None:
                sp = self.planet_spec(**pl)
            else:
                sp = stellar_spectrum(pl['sptype'])
            renorm_args = pl['renorm_args']
            if (renorm_args is not None) and (len(renorm_args) > 0):
                sp_norm = sp.renorm(*renorm_args)
                sp_norm.name = sp.name
                sp = sp_norm

            # Location relative to star
            xoff, yoff = pl['xyoff_pix']
            
            # Add in PA offset
            if PA_offset!=0:
                xoff, yoff = xy_rot(xoff, yoff, PA_offset)
            # Convert to arcsec
            xoff_asec, yoff_asec = np.array([xoff, yoff]) * self.pix_scale

            # Add in bar offset for PSF generation
            xoff_asec += self.bar_offset
            r, th = xy_to_rtheta(xoff_asec, yoff_asec)
            if quick_PSF:
                psf_planet = self.gen_offset_psf(r, th, return_oversample=False)
                obs = S.Observation(sp, self.bandpass, binset=self.bandpass.wave)
                psf_planet *= obs.effstim('counts')
            else:
                psf_planet = self.gen_offset_psf(r, th, sp=sp, return_oversample=False)
                
            # Expand to full size
            psf_planet = pad_or_cut_to_size(psf_planet, image_shape)
            # Shift to position relative to center of image
            delx, dely = (xoff + bar_offpix, yoff)
            if ('FULL' in self.det_info['wind_mode']) and (self.mask is not None):
                cdict = coron_ap_locs(self.module, self.channel, self.mask, full=True)
                xcen, ycen = cdict['cen_V23']
                delx += (xcen - xpix/2)
                dely += (ycen - ypix/2)
            psf_planet = fshift(psf_planet, delx=delx, dely=dely, pad=True)
            
            # Determine if any throughput loss due to coronagraphic mask 
            # artifacts, such as the mask holder or ND squares.
            # Planet positions are relative to the center of the mask,
            # which is not centered in a full detector. 
            # All subarrays should have the mask placed at the center.
            detid = self.Detectors[0].detid
            cmask = self.mask_images[detid]
            if use_cmask and (cmask is not None):
                # First, anything in a rectangular region around the 
                # mask has already been correctly accounted for
                if (np.abs(xoff_asec+bar_offset)<10) and (np.abs(yoff_asec)<5):
                    trans = 1
                elif 'FULL' in self.det_info['wind_mode']:
                    # If a full detector observation, then adjust
                    # to be relative to mask location
                    xpos = int(xcen + xoff + bar_offpix)
                    ypos = int(ycen + yoff)
                    cmask_sub = cmask[ypos-3:ypos+3,xpos-3:xpos+3]
                    trans = np.mean(cmask_sub)
                else:
                    xpos, ypox = (int(delx), int(dely))
                    cmask_sub = cmask[ypos-3:ypos+3,xpos-3:xpos+3]
                    trans = np.mean(cmask_sub)
                    
                #print(trans)
                psf_planet *= trans
            
            # Add to image
            image += psf_planet
            
        return image
        
    def kill_planets(self):
        """Remove planet info"""
        self._planets = []
    
    
    def gen_disk_image(self, PA_offset=0, use_cmask=False, **kwargs):
        """Create image of just disk.
        
        Generate a (noiseless) convolved image of the disk at some PA offset. 
        The PA offset value will rotate the image CCW. 
        Image units of e-/sec.
        
        Coordinate convention is for +V3 up and +V2 to left.
        
        Parameters
        ----------        
        PA_offset : float
            Rotate entire scene by some position angle.
            Positive values are counter-clockwise from +Y direction.
            Corresponds to instrument aperture PA.
        use_cmask : bool
            Use the coronagraphic mask image to attenuate disk regions
            getting obscurred by a corongraphic mask feature
        """
            
        if self.disk_hdulist is None:
            return 0.0

        # Final image shape
        image_shape = ypix, xpix = (self.det_info['ypix'], self.det_info['xpix'])
        bar_offset = self.bar_offset
        bar_offpix = bar_offset / self.pixelscale
        
        # The detector-sampled disk image
        disk_image  = self.disk_hdulist[0].data
        header = self.disk_hdulist[0].header
        if PA_offset!=0:
            disk_image = rotate(disk_image, -PA_offset, reshape=False, order=1)
            
        # Shift rotated image to location along bar
        delx, dely = (bar_offpix, 0)
        # Shift to position relative to center of image
        if ('FULL' in self.det_info['wind_mode']) and (self.mask is not None):
            cdict = coron_ap_locs(self.module, self.channel, self.mask, full=True)
            xcen, ycen = cdict['cen_V23']
        else:
            xcen, ycen = (xpix/2, ypix/2)
        delx += (xcen - xpix/2)
        dely += (ycen - ypix/2)
        
        disk_image = fshift(disk_image, delx=delx, dely=dely, pad=True)

        # Multiply raw disk data by coronagraphic mask. 
        # Exclude region already affected by observed mask.
        detid = self.Detectors[0].detid
        cmask = self.mask_images[detid]
        if use_cmask and (cmask is not None):
            # Exclude actual coronagraphic mask since this will be
            # taken into account during PSF convolution. Not true for
            # all other elements within FOV, ND squares, and mask holder.
            cmask_temp = cmask.copy()
            if 'FULL' in self.det_info['wind_mode']:
                #cdict = coron_ap_locs(self.module, self.channel, self.mask, full=True)
                #xcen, ycen = cdict['cen_V23']
                r, th = dist_image(cmask, pixscale=self.pixelscale, 
                                   center=cdict['cen_V23'], return_theta=True)
                x_asec, y_asec = rtheta_to_xy(r, th)
                ind = (np.abs(x_asec)<10) & (np.abs(y_asec)<5)
                cmask_temp[ind] = 1
            else:
                r, th = dist_image(cmask, pixscale=self.pixelscale, return_theta=True)
                x_asec, y_asec = rtheta_to_xy(r, th)
                ind = (np.abs(x_asec)<10) & (np.abs(y_asec)<5)
                cmask_temp[ind] = 1

            disk_image *= cmask_temp

        ##################################
        # Image convolution
        ##################################
        
        noff = len(self.offset_list)
        xypix = int(np.sqrt(ypix*xpix))
        nproc = nproc_use_convolve(xypix, 1, noff)
            
        if (noff==1): # Single PSF
            psf = self.psf_list[0]
            # Normalize PSF sum to 1.0
            # Otherwise convolve_fft may throw an error if psf.sum() is too small
            norm = psf.sum()
            psf = psf / norm
            image_conv = convolve_fft(disk_image, psf, fftn=fftpack.fftn, 
                                      ifftn=fftpack.ifftn, allow_huge=True)
            image_conv *= norm

        else:
            r, th = dist_image(cmask, pixscale=self.pixelscale, 
                               center=(xcen,ycen), return_theta=True)
            x_asec, y_asec = rtheta_to_xy(r, th)
            ind = (np.abs(x_asec)<10) & (np.abs(y_asec)<5)
            
            # Remove ND squares from mask for convolution purposes
            cmask_temp = cmask.copy()
            cmask_temp[~ind] = 1
                    
            #if self.mask[-1]=='B': # Bar mask
            # Create a mask the size/shape of PSF
            fov_pix  = self.psf_info['fov_pix']
            fov_asec = fov_pix * self.pixelscale
            im_mask = coron_trans(self.mask, fov=fov_asec, 
                                  pixscale=self.pixelscale, nd_squares=False)
            im_mask = pad_or_cut_to_size(im_mask, (fov_pix,fov_pix))
            mask_cut = im_mask[:,int(fov_pix/2)]
            yarr = (np.arange(fov_pix) - fov_pix/2) * self.pixelscale

            # Mask transmission value ranges associated with 
            # each psf in self.psf_list. These were generated
            # by offsetting vertically from center of mask.
            tvals = np.interp(self.offset_list, yarr, mask_cut)
            
            # Resort tvals and self.psf_list by transmission 
            isort = np.argsort(tvals)
            psf_list_sort = [self.psf_list[i] for i in isort]
            
            tvals = tvals[isort]
            tvals_del = (tvals[1:] - tvals[0:-1])
            tvals_mid = tvals[0:-1] + tvals_del / 2
            tvals_edges = np.array([0] + list(tvals_mid) + [1])

            worker_args = [(psf, disk_image, tvals_edges, cmask_temp, i) 
                           for i,psf in enumerate(psf_list_sort)]

            if nproc<=1:
                imconv_slices = [_wrap_conv_trans_for_mp(wa) for wa in worker_args]
            else:
                pool = mp.Pool(nproc)
                try:
                    imconv_slices = pool.map(_wrap_conv_trans_for_mp, worker_args)
                except Exception as e:
                    print('Caught an exception during multiprocess:')
                    raise e
                finally:
                    pool.close()

            # Turn into a numpy array of shape (noff,nx,ny)
            imconv_slices = np.array(imconv_slices)
            # Sum all images together
            image_conv = imconv_slices.sum(axis=0)
            
        image_conv[image_conv<0] = 0
        return image_conv


    def star_flux(self, fluxunit='counts', sp=None):
        """ Stellar flux.

        Return the stellar flux in whatever units, such as 
        vegamag, counts, or Jy.
        
        Parameters
        ----------
        fluxunits : str
            Desired output units, such as counts, vegamag, Jy, etc.
            Must be a Pysynphot supported unit string.
        sp : :mod:`pysynphot.spectrum`
            Normalized Pysynphot spectrum.
        """
        
        if sp is None: sp=self.sp_sci

        # Create pysynphot observation
        bp = self.bandpass
        obs = S.Observation(sp, bp, binset=bp.wave)
        
        return obs.effstim(fluxunit)

    def _fix_sat_im(self, image, sat_val=0.8, **kwargs):
        """Fix saturated region of an image
        
        
        Parameters
        ----------
        image : ndarray
            Image to clean.
        sav_val : float
            Well level fraction to considered saturated.
                    
        Keyword Args
        ------------
        full_size : bool
            Expand (or contract) to size of detector array?
            If False, returned image is fov_pix size.
        ngroup : int
            How many group times to determine saturation level?
            If this number is higher than the total groups in ramp, 
            then a warning is produced. The default is ngroup=2, 
            A value of 0 corresponds to the so-called "zero-frame," 
            which is the very first frame that is read-out and saved 
            separately. This is the equivalent to ngroup=1 for RAPID
            and BRIGHT1 observations.
        do_ref : bool
            Get saturation levels for reference source instead of science.
        niter_max : int
            Number of iterations for fixing NaNs. Default=5.
        """

        sat_level = self.saturation_levels(self, image=image, **kwargs)
        sat_mask = sat_level > sat_val
        image[sat_mask] = np.nan
        image = fix_nans_with_med(image, **kwargs)
        
        # If there are any leftover NaNs, make them 0.
        nan_mask = np.isnan(image)
        image[nan_mask] = 0
        
        return image

    def _gen_roll_image_old(self, PA1=0, PA2=10, zfact=None, oversample=None, 
        exclude_disk=False, exclude_planets=False, exclude_noise=False, 
        no_ref=False, opt_diff=True, fix_sat=False, **kwargs):
        """Make roll-subtracted image.
        
        Create a final roll-subtracted slope image based on current observation
        settings. Coordinate convention is for +V3 up and +V2 to left.
        
        Procedure:
        
        - Create Roll 1 and Roll 2 slope images (star+exoplanets+disk)
        - Create Reference Star slope image
        - Add noise to all images
        - Subtract ref image from both rolls
        - De-rotate Roll 2 by roll_angle value
        - Average Roll 1 and de-rotated Roll 2
        - De-rotate averaged image to North

        Returns an HDUList of final image (North rotated upwards).

          
        Parameters
        ----------
        PA1 : float
            Position angle of first roll position (clockwise, from East to West)
        PA2 : float, None
            Position angle of second roll position. If set equal to PA1 
            (or to None), then only one roll will be performed.
            Otherwise, two rolls are performed, each using the specified 
            MULTIACCUM settings (doubling the effective exposure time).
        zfact : float
            Zodiacal background factor (default=2.5)
        oversample : float
            Set oversampling of final image.
        exclude_disk : bool
            Ignore disk when subtracted image (for radial contrast),
            but still add Poisson noise from disk.
        exclude_noise : bool
            Don't add random Gaussian noise (detector+photon)
        no_ref : bool
            Exclude reference observation. Subtraction is then Roll1-Roll2.
        opt_diff : bool
            Optimal reference differencing (scaling only on the inner regions)
        fix_sat : bool
            Calculate saturated regions and fix with median of nearby data.
            
        Keyword Args
        ------------
        quick_PSF : bool
            Rather than generate a spectrum-weighted PSF for planets, use 
            the cached PSF scaled by the photon count through the bandpass.
            Resulting PSFs are slightly less accurate, but much faster.
            Default is True.
        use_cmask : bool
            Use the coronagraphic mask image to attenuate planet or disk that
            is obscurred by a corongraphic mask feature.

        """
    
        # Final image shape
        image_shape = xpix, ypix = (self.det_info['xpix'], self.det_info['ypix'])
        # Sub-image for determining ref star scale factor
        subsize = 50
        xsub = np.min([subsize,xpix])
        ysub = np.min([subsize,ypix])
        sub_shape = (ysub, xsub)
        
        # Position angle decisions
        if PA2 is None: 
            roll_angle = 0
        else:
            roll_angle = PA2 - PA1
        if oversample is None: oversample = 1

        # Center location of star
        bar_offpix = self.bar_offset / self.pixelscale
        if ('FULL' in self.det_info['wind_mode']) and (self.mask is not None):
            cdict = coron_ap_locs(self.module, self.channel, self.mask, full=True)
            xcen, ycen = cdict['cen_V23']
            xcen += bar_offpix
        else:
            xcen, ycen = (xpix/2 + bar_offpix, ypix/2)
        delx, dely = (xcen - xpix/2, ycen - ypix/2)

        # Account for possible oversampling
        xcen_over, ycen_over = np.array([xcen, ycen]) * oversample
        delx_over, dely_over = np.array([delx, dely]) * oversample
        cen_over = (xcen_over, ycen_over)

        
        if no_ref and (roll_angle==0):
            _log.warning('If no_ref=True, then PA1 must not equal PA2. Setting no_ref=False')
            no_ref = False
   
        sci = self
        ref = self.nrc_ref
        
        # Stellar PSF is fixed (doesn't rotate)
        im_star = sci.gen_psf(self.sp_sci, return_oversample=False)
        im_star_sub = pad_or_cut_to_size(im_star, sub_shape)
        im_star = pad_or_cut_to_size(im_star, image_shape)
        im_star = fshift(im_star, delx=delx, dely=dely, pad=True)
        
        # Disk and Planet images
        no_disk    = exclude_disk    and exclude_noise
        no_planets = exclude_planets and exclude_noise
        im_disk_r1 = 0 if no_disk    else sci.gen_disk_image(PA_offset=PA1, **kwargs)
        im_pl_r1   = 0 if no_planets else sci.gen_planets_image(PA_offset=PA1, **kwargs)

        # Telescope Roll 1
        im_roll1 = im_star + im_disk_r1 + im_pl_r1
        # Noise per pixel
        if not exclude_noise:
            det = sci.Detectors[0]
            fzodi = sci.bg_zodi(zfact)
            im_noise1 = det.pixel_noise(fsrc=im_roll1, fzodi=fzodi)
            # Add random noise
            im_roll1 += np.random.normal(scale=im_noise1)
            
        # Roll1
        # Get rid of disk and planet emission
        # but we want to keep their noise contribution
        if exclude_disk: im_roll1 -= im_disk_r1
        if exclude_planets: im_roll1 -= im_pl_r1

        if fix_sat:  # Fix saturated pixels
            im_roll1 = self._fix_sat_im(im_roll1, **kwargs)

        # Pure roll subtraction (no reference PSF)
        ##################################################
        
        if no_ref: 
            # Roll2
            if self.wfe_roll_drift>0:
                # Change self.wfe_drift to self.wfe_roll_drift
                # and calculate new stellar PSF for Roll2
                self._set_wfe_drift(self.wfe_roll_drift)
                im_star2 = sci.gen_psf(sci.sp_sci, return_oversample=False)
                im_star2 = pad_or_cut_to_size(im_star2, image_shape)
                im_star2 = fshift(im_star, delx=delx, dely=dely, pad=True)
                self._set_wfe_drift(0)
            else:
                im_star2 = im_star
            
            im_disk_r2 = 0 if no_disk else sci.gen_disk_image(PA_offset=PA2, **kwargs)
            im_pl_r2   = 0 if no_planets else sci.gen_planets_image(PA_offset=PA2, **kwargs)
            im_roll2   = im_star2 + im_disk_r2 + im_pl_r2

            # Noise per pixel
            if not exclude_noise:
                det = sci.Detectors[0]
                fzodi = sci.bg_zodi(zfact)
                im_noise2 = det.pixel_noise(fsrc=im_roll2, fzodi=fzodi)
                # Add random noise
                im_roll2 += np.random.normal(scale=im_noise2)

            if exclude_disk: im_roll2 -= im_disk_r2
            if exclude_planets: im_roll2 -= im_pl_r2

            if fix_sat:  # Fix saturated pixels
                im_roll2 = self._fix_sat_im(im_roll2, **kwargs)

            if oversample != 1:
                im_roll1 = frebin(im_roll1, scale=oversample)
                im_roll2 = frebin(im_roll2, scale=oversample)

            diff_r1 = im_roll1 - im_roll2
            diff_r2 = -1 * diff_r1
            
            # De-rotate each image
            diff_r1_rot = rotate_offset(diff_r1, PA1, cen=cen_over, reshape=True, cval=np.nan)
            diff_r2_rot = rotate_offset(diff_r2, PA2, cen=cen_over, reshape=True, cval=np.nan)

            # Expand to the same size
            new_shape = tuple(np.max(np.array([diff_r1_rot.shape, diff_r2_rot.shape]), axis=0))
            diff_r1_rot = pad_or_cut_to_size(diff_r1_rot, new_shape, np.nan)
            diff_r2_rot = pad_or_cut_to_size(diff_r2_rot, new_shape, np.nan)

            # Replace NaNs with values from other differenced mask
            nan_mask = np.isnan(diff_r1_rot)
            diff_r1_rot[nan_mask] = diff_r2_rot[nan_mask]
            nan_mask = np.isnan(diff_r2_rot)
            diff_r2_rot[nan_mask] = diff_r1_rot[nan_mask]

            final = (diff_r1_rot + diff_r2_rot) / 2
                
            hdu = fits.PrimaryHDU(final)
            hdu.header['EXTNAME'] = ('ROLL_SUB')
            hdu.header['OVERSAMP'] = oversample
            hdu.header['PIXELSCL'] = sci.pix_scale / hdu.header['OVERSAMP']
            hdulist = fits.HDUList([hdu])

            return hdulist

        
        # Continuing with a ref PSF subtraction algorithm
        ##################################################
        
        # Reference star slope simulation
        # Ideal slope
        im_ref = ref.gen_psf(self.sp_ref, return_oversample=False)
        im_ref_sub = pad_or_cut_to_size(im_ref, sub_shape)
        im_ref = pad_or_cut_to_size(im_ref, image_shape)
        im_ref = fshift(im_ref, delx=delx, dely=dely, pad=True)
        # Noise per pixel
        if not exclude_noise:
            det = ref.Detectors[0]
            fzodi = ref.bg_zodi(zfact)
            im_noise = det.pixel_noise(fsrc=im_ref, fzodi=fzodi)
            # Add random noise
            im_ref += np.random.normal(scale=im_noise)

        if fix_sat:  # Fix saturated pixels
            im_ref = self._fix_sat_im(im_ref, **kwargs)

        # Determine reference star scale factor
        scale1 = scale_ref_image(im_star_sub, im_ref_sub)
        _log.debug('scale1: {0:.3f}'.format(scale1))
        if oversample != 1:
            im_ref_rebin = frebin(im_ref, scale=oversample)
            im_roll1     = frebin(im_roll1, scale=oversample)
        else:
            im_ref_rebin = im_ref
            
        # Telescope Roll 2 with reference subtraction
        if (abs(roll_angle) > eps) and (not no_ref):
            # Subtraction with and without scaling
            im_diff1_r1 = im_roll1 - im_ref_rebin
            im_diff2_r1 = im_roll1 - im_ref_rebin * scale1

            if self.wfe_roll_drift>0:
                self._set_wfe_drift(self.wfe_roll_drift)
                im_star2 = sci.gen_psf(sci.sp_sci, return_oversample=False)
                im_star2_sub = pad_or_cut_to_size(im_star2, sub_shape)
                im_star2 = pad_or_cut_to_size(im_star2, image_shape)
                im_star2 = fshift(im_star2, delx=delx, dely=dely, pad=True)
                self._set_wfe_drift(0)
            else:
                im_star2 = im_star
                im_star2_sub = im_star_sub
                
            im_disk_r2 = sci.gen_disk_image(PA_offset=PA2, **kwargs)
            im_pl_r2   = sci.gen_planets_image(PA_offset=PA2, **kwargs)
            im_roll2   = im_star2 + im_disk_r2 + im_pl_r2
            # Noise per pixel
            if not exclude_noise:
                det = sci.Detectors[0]
                fzodi = sci.bg_zodi(zfact)
                im_noise2 = det.pixel_noise(fsrc=im_roll2, fzodi=fzodi)
                # Add random noise
                im_roll2 += np.random.normal(scale=im_noise2)

            if exclude_disk:
                im_roll2 -= im_disk_r2
            if exclude_planets:
                im_roll2 -= im_pl_r2

            if fix_sat:  # Fix saturated pixels
                im_roll2 = self._fix_sat_im(im_roll2, **kwargs)

            # Subtract reference star from Roll 2
            #im_star_sub = pad_or_cut_to_size(im_star+im_pl_r2, sub_shape)
            scale2 = scale_ref_image(im_star2_sub, im_ref_sub)
            _log.debug('scale2: {0:.3f}'.format(scale2))
            if oversample != 1:
                im_roll2 = frebin(im_roll2, scale=oversample)
            # Subtraction with and without scaling
            im_diff1_r2 = im_roll2 - im_ref_rebin
            im_diff2_r2 = im_roll2 - im_ref_rebin * scale2
            #im_diff_r2 = optimal_difference(im_roll2, im_ref_rebin, scale2)

            # De-rotate each image
            # Convention for rotate() is opposite PA_offset
            diff1_r1_rot = rotate_offset(im_diff1_r1, PA1, cen=cen_over, reshape=True, cval=np.nan)
            diff2_r1_rot = rotate_offset(im_diff2_r1, PA1, cen=cen_over, reshape=True, cval=np.nan)
            diff1_r2_rot = rotate_offset(im_diff1_r2, PA2, cen=cen_over, reshape=True, cval=np.nan)
            diff2_r2_rot = rotate_offset(im_diff2_r2, PA2, cen=cen_over, reshape=True, cval=np.nan)

            # Expand all images to the same size
            new_shape = tuple(np.max(np.array([diff1_r1_rot.shape, diff1_r2_rot.shape]), axis=0))
            diff1_r1_rot = pad_or_cut_to_size(diff1_r1_rot, new_shape, np.nan)
            diff2_r1_rot = pad_or_cut_to_size(diff2_r1_rot, new_shape, np.nan)
            diff1_r2_rot = pad_or_cut_to_size(diff1_r2_rot, new_shape, np.nan)
            diff2_r2_rot = pad_or_cut_to_size(diff2_r2_rot, new_shape, np.nan)
            
            # Replace NaNs with values from other differenced mask
            nan_mask = np.isnan(diff1_r1_rot)
            diff1_r1_rot[nan_mask] = diff1_r2_rot[nan_mask]
            diff2_r1_rot[nan_mask] = diff2_r2_rot[nan_mask]
            nan_mask = np.isnan(diff1_r2_rot)
            diff1_r2_rot[nan_mask] = diff1_r1_rot[nan_mask]
            diff2_r2_rot[nan_mask] = diff2_r1_rot[nan_mask]
            
            # final1 has better noise in outer regions (background)
            # final2 has better noise in inner regions (PSF removal)
            final1 = (diff1_r1_rot + diff1_r2_rot) / 2
            final2 = (diff2_r1_rot + diff2_r2_rot) / 2
            
            if opt_diff:
                rho = dist_image(final1)
                binsize = 1
                bins = np.arange(rho.min(), rho.max() + binsize, binsize)

                nan_mask = np.isnan(final1)
                igroups = hist_indices(rho[~nan_mask], bins)

                std1 = binned_statistic(igroups, final1[~nan_mask], func=np.std)
                std2 = binned_statistic(igroups, final2[~nan_mask], func=np.std)
                
                ibin_better = np.where(std1 < std2)[0]
                for ibin in ibin_better:
                    final2.ravel()[igroups[ibin]] = final1.ravel()[igroups[ibin]]
                    
            final = final2

        else:
            # Optimal differencing (with scaling only on the inner regions)
            if opt_diff:
                final = optimal_difference(im_roll1, im_ref_rebin, scale1)
            else:
                final = im_roll1 - im_ref_rebin * scale1
                
            final = rotate_offset(final, PA1, cen=cen_over, reshape=True, cval=np.nan)
            
        # De-rotate PA1 to North
        #if abs(PA1) > eps:
        #    final = rotate(final, PA1, reshape=False)
        
        hdu = fits.PrimaryHDU(final)
        hdu.header['EXTNAME'] = ('ROLL_SUB')
        hdu.header['OVERSAMP'] = oversample
        hdu.header['PIXELSCL'] = sci.pix_scale / hdu.header['OVERSAMP']
        hdulist = fits.HDUList([hdu])

        return hdulist

    def gen_roll_image(self, PA1=0, PA2=10, zfact=None, oversample=None, 
        no_ref=False, opt_diff=True, fix_sat=False, **kwargs):
        """Make roll-subtracted image.
        
        Create a final roll-subtracted slope image based on current observation
        settings. Coordinate convention is for +V3 up and +V2 to left.
        
        Procedure:
        
        - Create Roll 1 and Roll 2 slope images (star+exoplanets+disk)
        - Create Reference Star slope image
        - Add noise to all images
        - Scale ref image
        - Subtract ref image from both rolls
        - De-rotate Roll 1 and Roll 2 to common V2/V3
        - Average Roll 1 and Roll 2

        Returns an HDUList of final image (V3 rotated upwards).
         
        Parameters
        ----------
        PA1 : float
            Position angle of first roll position (clockwise, from East to West)
        PA2 : float, None
            Position angle of second roll position. If set equal to PA1 
            (or to None), then only one roll will be performed.
            Otherwise, two rolls are performed, each using the specified 
            MULTIACCUM settings (doubling the effective exposure time).
        oversample : float
            Set oversampling of final image.
        no_ref : bool
            Exclude reference observation. Subtraction is then Roll1-Roll2.
        opt_diff : bool
            Optimal reference differencing (scaling only on the inner regions)
        fix_sat : bool
            Calculate saturated regions and fix with median of nearby data.
            
        Keyword Args
        ------------
        exclude_disk : bool
            Do not include disk in final image (for radial contrast),
            but still add Poisson noise from disk.
        exclude_planets : bool
            Do not include planets in final image (for radial contrast),
            but still add Poisson noise from disk.
        exclude_noise : bool
            Don't add random Gaussian noise (detector+photon)
        quick_PSF : bool
            Rather than generate a spectrum-weighted PSF for planets, use 
            the cached PSF scaled by the photon count through the bandpass.
            Resulting PSFs are slightly less accurate, but much faster.
            Default is True.
        use_cmask : bool
            Use the coronagraphic mask image to attenuate planet or disk that
            is obscurred by a corongraphic mask feature.
        zfact : float
            Zodiacal background factor (default=2.5)
        locstr : 
            Object name or RA/DEC (decimal degrees or sexigesimal)
        year : int
            Year of observation
        day : float
            Day of observation

        """
    
        # Final image shape
        image_shape = xpix, ypix = (self.det_info['xpix'], self.det_info['ypix'])
        # Sub-image for determining ref star scale factor
        subsize = 50
        xsub = np.min([subsize,xpix])
        ysub = np.min([subsize,ypix])
        sub_shape = (ysub, xsub)
        
        # Position angle decisions
        if PA2 is None: 
            roll_angle = 0
        else:
            roll_angle = PA2 - PA1
        if oversample is None: oversample = 1

        # Center location of star
        bar_offpix = self.bar_offset / self.pixelscale
        if ('FULL' in self.det_info['wind_mode']) and (self.mask is not None):
            cdict = coron_ap_locs(self.module, self.channel, self.mask, full=True)
            xcen, ycen = cdict['cen_V23']
            xcen += bar_offpix
        else:
            xcen, ycen = (xpix/2 + bar_offpix, ypix/2)
        delx, dely = (xcen - xpix/2, ycen - ypix/2)

        # Account for possible oversampling
        xcen_over, ycen_over = np.array([xcen, ycen]) * oversample
        delx_over, dely_over = np.array([delx, dely]) * oversample
        cen_over = (xcen_over, ycen_over)

        
        if no_ref and (roll_angle==0):
            _log.warning('If no_ref=True, then PA1 must not equal PA2. Setting no_ref=False')
            no_ref = False
   
        sci = self
        ref = self.nrc_ref
        
        # Stellar PSF is fixed (doesn't rotate)
        im_star = sci.gen_psf(self.sp_sci, return_oversample=False)
        im_star_sub = pad_or_cut_to_size(im_star, sub_shape)
        im_star = pad_or_cut_to_size(im_star, image_shape)
        im_star = fshift(im_star, delx=delx, dely=dely, pad=True)
                
        im_roll1 = self.gen_slope_image(PA=PA1, im_star=im_star, **kwargs)
        
        # Fix saturated pixels
        if fix_sat:  
            im_roll1 = self._fix_sat_im(im_roll1, **kwargs)

        # Pure roll subtraction (no reference PSF)
        ##################################################
        
        if no_ref: 
            # Roll2
            wfe_drift1 = self.wfe_drift
            wfe_drift2 = wfe_drift1 + self.wfe_roll_drift
            
            # Change self.wfe_drift, gen image, and return wfe_drift
            if np.abs(self.wfe_roll_drift) > eps:
                im_roll2 = self.gen_slope_image(PA=PA2, im_star=im_star, **kwargs)
            else:
                self._set_wfe_drift(wfe_drift2)
                im_roll2 = self.gen_slope_image(PA=PA2, **kwargs)
                self._set_wfe_drift(wfe_drift1)
            
            # Fix saturated pixels
            if fix_sat:  
                im_roll2 = self._fix_sat_im(im_roll2, **kwargs)

            if oversample != 1:
                im_roll1 = frebin(im_roll1, scale=oversample)
                im_roll2 = frebin(im_roll2, scale=oversample)

            diff_r1 = im_roll1 - im_roll2
            diff_r2 = -1 * diff_r1
            
            # De-rotate each image
            diff_r1_rot = rotate_offset(diff_r1, PA1, cen=cen_over, reshape=True, cval=np.nan)
            diff_r2_rot = rotate_offset(diff_r2, PA2, cen=cen_over, reshape=True, cval=np.nan)

            # Expand to the same size
            new_shape = tuple(np.max(np.array([diff_r1_rot.shape, diff_r2_rot.shape]), axis=0))
            diff_r1_rot = pad_or_cut_to_size(diff_r1_rot, new_shape, np.nan)
            diff_r2_rot = pad_or_cut_to_size(diff_r2_rot, new_shape, np.nan)

            # Replace NaNs with values from other differenced mask
            nan_mask = np.isnan(diff_r1_rot)
            diff_r1_rot[nan_mask] = diff_r2_rot[nan_mask]
            nan_mask = np.isnan(diff_r2_rot)
            diff_r2_rot[nan_mask] = diff_r1_rot[nan_mask]

            final = (diff_r1_rot + diff_r2_rot) / 2
                
            hdu = fits.PrimaryHDU(final)
            hdu.header['EXTNAME'] = ('ROLL_SUB')
            hdu.header['OVERSAMP'] = oversample
            hdu.header['PIXELSCL'] = sci.pix_scale / hdu.header['OVERSAMP']
            hdulist = fits.HDUList([hdu])

            return hdulist
        
        # Continuing with a ref PSF subtraction algorithm
        ##################################################
        
        # Reference star slope simulation
        # Ideal slope
        im_ref = ref.gen_psf(self.sp_ref, return_oversample=False)
        im_ref_sub = pad_or_cut_to_size(im_ref, sub_shape)
        im_ref = pad_or_cut_to_size(im_ref, image_shape)
        im_ref = fshift(im_ref, delx=delx, dely=dely, pad=True)
        
        # With noise
        im_ref = self.gen_slope_image(do_ref=True, im_star=im_ref, **kwargs)

        # Fix saturated pixels
        if fix_sat:  
            im_ref = self._fix_sat_im(im_ref, do_ref=True, **kwargs)

        # Determine reference star scale factor
        scale1 = scale_ref_image(im_star_sub, im_ref_sub)
        _log.debug('scale1: {0:.3f}'.format(scale1))
        if oversample != 1:
            im_ref   = frebin(im_ref, scale=oversample)
            im_roll1 = frebin(im_roll1, scale=oversample)
            
        # Telescope Roll 2 with reference subtraction
        if (abs(roll_angle) > eps):
            # Subtraction with and without scaling
            im_diff1_r1 = im_roll1 - im_ref
            im_diff2_r1 = im_roll1 - im_ref * scale1

            # WFE drift difference between rolls
            wfe_drift1 = self.wfe_drift
            wfe_drift2 = wfe_drift1 + self.wfe_roll_drift
            
            # Change self.wfe_drift, gen image, and return wfe_drift
            self._set_wfe_drift(wfe_drift2)
            if np.abs(self.wfe_roll_drift) > eps:
                im_star2 = sci.gen_psf(sci.sp_sci, return_oversample=False)
                im_star2_sub = pad_or_cut_to_size(im_star2, sub_shape)
                im_star2 = pad_or_cut_to_size(im_star2, image_shape)
                im_star2 = fshift(im_star2, delx=delx, dely=dely, pad=True)
            else:
                im_star2 = im_star
                im_star2_sub = im_star_sub
            im_roll2 = self.gen_slope_image(PA=PA2, im_star=im_star2, **kwargs)
            self._set_wfe_drift(wfe_drift1)

            # Fix saturated pixels
            if fix_sat:  
                im_roll2 = self._fix_sat_im(im_roll2, **kwargs)

            # Subtract reference star from Roll 2
            #im_star_sub = pad_or_cut_to_size(im_star+im_pl_r2, sub_shape)
            scale2 = scale_ref_image(im_star2_sub, im_ref_sub)
            _log.debug('scale2: {0:.3f}'.format(scale2))
            if oversample != 1:
                im_roll2 = frebin(im_roll2, scale=oversample)
            # Subtraction with and without scaling
            im_diff1_r2 = im_roll2 - im_ref
            im_diff2_r2 = im_roll2 - im_ref * scale2
            #im_diff_r2 = optimal_difference(im_roll2, im_ref, scale2)

            # De-rotate each image
            # Convention for rotate() is opposite PA_offset
            diff1_r1_rot = rotate_offset(im_diff1_r1, PA1, cen=cen_over, reshape=True, cval=np.nan)
            diff2_r1_rot = rotate_offset(im_diff2_r1, PA1, cen=cen_over, reshape=True, cval=np.nan)
            diff1_r2_rot = rotate_offset(im_diff1_r2, PA2, cen=cen_over, reshape=True, cval=np.nan)
            diff2_r2_rot = rotate_offset(im_diff2_r2, PA2, cen=cen_over, reshape=True, cval=np.nan)

            # Expand all images to the same size
            new_shape = tuple(np.max(np.array([diff1_r1_rot.shape, diff1_r2_rot.shape]), axis=0))
            diff1_r1_rot = pad_or_cut_to_size(diff1_r1_rot, new_shape, np.nan)
            diff2_r1_rot = pad_or_cut_to_size(diff2_r1_rot, new_shape, np.nan)
            diff1_r2_rot = pad_or_cut_to_size(diff1_r2_rot, new_shape, np.nan)
            diff2_r2_rot = pad_or_cut_to_size(diff2_r2_rot, new_shape, np.nan)
            
            # Replace NaNs with values from other differenced mask
            nan_mask = np.isnan(diff1_r1_rot)
            diff1_r1_rot[nan_mask] = diff1_r2_rot[nan_mask]
            diff2_r1_rot[nan_mask] = diff2_r2_rot[nan_mask]
            nan_mask = np.isnan(diff1_r2_rot)
            diff1_r2_rot[nan_mask] = diff1_r1_rot[nan_mask]
            diff2_r2_rot[nan_mask] = diff2_r1_rot[nan_mask]
            
            # final1 has better noise in outer regions (background)
            # final2 has better noise in inner regions (PSF removal)
            final1 = (diff1_r1_rot + diff1_r2_rot) / 2
            final2 = (diff2_r1_rot + diff2_r2_rot) / 2
            
            if opt_diff:
                rho = dist_image(final1)
                binsize = 1
                bins = np.arange(rho.min(), rho.max() + binsize, binsize)

                nan_mask = np.isnan(final1)
                igroups = hist_indices(rho[~nan_mask], bins)

                std1 = binned_statistic(igroups, final1[~nan_mask], func=np.std)
                std2 = binned_statistic(igroups, final2[~nan_mask], func=np.std)
                
                ibin_better = np.where(std1 < std2)[0]
                for ibin in ibin_better:
                    final2.ravel()[igroups[ibin]] = final1.ravel()[igroups[ibin]]
                    
            final = final2

        # For only a single roll
        else:
            # Optimal differencing (with scaling only on the inner regions)
            if opt_diff:
                final = optimal_difference(im_roll1, im_ref, scale1)
            else:
                final = im_roll1 - im_ref * scale1
                
            final = rotate_offset(final, PA1, cen=cen_over, reshape=True, cval=np.nan)
            
        # De-rotate PA1 to North
        #if abs(PA1) > eps:
        #    final = rotate(final, PA1, reshape=False)
        
        hdu = fits.PrimaryHDU(final)
        hdu.header['EXTNAME'] = ('ROLL_SUB')
        hdu.header['OVERSAMP'] = oversample
        hdu.header['PIXELSCL'] = sci.pix_scale / hdu.header['OVERSAMP']
        hdulist = fits.HDUList([hdu])

        return hdulist

    def calc_contrast(self, hdu_diff=None, roll_angle=10, nsig=1, 
        exclude_disk=True, exclude_planets=True, no_ref=False, **kwargs):
        """Create contrast curve.
        
        Generate n-sigma contrast curve for the current observation settings.
        Make sure that MULTIACCUM parameters are set for both the main
        class (``self.update_detectors()``) as well as the reference target
        class (``self.nrc_ref.update_detectors()``).
        
        Parameters
        ----------
        hdu_diff : HDUList
            Option to pass an already pre-made difference image.
        
        roll_angle : float
            Telescope roll angle (deg) between two observations.
            If set to 0 or None, then only one roll will be performed.
            If value is >0, then two rolls are performed, each using the
            specified MULTIACCUM settings (doubling the effective exposure
            time).
        nsig  : float
            n-sigma contrast curve.
        exclude_disk : bool
            Ignore disk when generating image?
        exclude_planets : bool
            Ignore planets when generating image?
        no_ref : bool
            Exclude reference observation. Subtraction is then Roll1-Roll2.
        
        Keyword Args
        ------------
        zfact : float
            Zodiacal background factor (default=2.5)
        exclude_noise : bool
            Don't add random Gaussian noise (detector+photon)?
        opt_diff : bool
            Optimal reference differencing (scaling only on the inner regions)

        Returns
        -------
        tuple
            Three arrays in a tuple: the radius in arcsec, n-sigma contrast,
            and n-sigma magnitude sensitivity limit (vega mag). 
        """
        if no_ref and (roll_angle==0):
            _log.warning('If no_ref=True, roll_angle must not equal 0. Setting no_ref=False')
            no_ref = False

        # If no HDUList is passed, then create one
        if hdu_diff is None:
            PA1 = 0
            PA2 = None if abs(roll_angle) < eps else roll_angle
            hdu_diff = self.gen_roll_image(PA1=PA1, PA2=PA2, exclude_disk=exclude_disk, 
                                           exclude_planets=exclude_planets, 
                                           no_ref=no_ref, **kwargs)
        
        # Radial noise
        data = hdu_diff[0].data
        header = hdu_diff[0].header
        
        data_rebin = frebin(data, scale=1/header['OVERSAMP'])
        pixscale = self.pixelscale
        rho = dist_image(data_rebin, pixscale=pixscale)

        # Get radial profiles
        binsize = pixscale
        bins = np.arange(rho.min(), rho.max() + binsize, binsize)
        nan_mask = np.isnan(data_rebin)
        igroups, _, rr = hist_indices(rho[~nan_mask], bins, True)
        stds = binned_statistic(igroups, data_rebin[~nan_mask], func=robust.medabsdev)
        stds = convolve(stds, Gaussian1DKernel(1))

        # Ignore corner regions
        xpix, ypix = (self.det_info['xpix'], self.det_info['ypix'])
        arr_size = np.min(data_rebin.shape) * pixscale
        mask = rr < (arr_size/2)
        rr = rr[mask]
        stds = stds[mask]

        # Normalize by psf max value
        if no_ref: 
            # No reference image subtraction; pure roll subtraction
            
            off_vals = []
            max_vals = []
            xvals = np.arange(1,xpix/2,5)
            for j, roff in enumerate(xvals):
                psf1 = self.gen_offset_psf(roff, 0, return_oversample=False)
                psf2 = self.gen_offset_psf(roff, roll_angle, return_oversample=False)
                
                psf1 = fshift(psf1, delx=0, dely=roff, pad=False)
                xoff, yoff = xy_rot(0, roff, 10)
                psf2 = fshift(psf2, delx=xoff, dely=yoff, pad=False)
                
                diff = psf1 - psf2
                maxv = diff.max()
                
                off_vals.append(roff)
                max_vals.append(maxv)
                if maxv >= 0.95*psf1.max():
                    off_vals = [0] + off_vals + [roff+5, xpix/2]
                    max_vals = [0] + max_vals + [psf1.max(), psf1.max()]
                    break
                    
            max_vals = np.array(max_vals)
            off_asec = np.array(off_vals) * pixscale
                
            psf_max = np.interp(rr, off_asec, max_vals)
                
        elif self.mask is None: # Direct imaging
            psf = self.gen_offset_psf(0, 0)
            psf_max = psf.max()
        elif self.mask[-1]=='R': # Round masks
            fov_asec = np.max([xpix,ypix]) * pixscale

            # Image mask
            im_mask = coron_trans(self.mask, fov=fov_asec, pixscale=pixscale, nd_squares=False)
            im_mask = pad_or_cut_to_size(im_mask, data_rebin.shape)

            ny, nx = im_mask.shape
            xv = (np.arange(nx) - nx/2) * pixscale
            
            # a and b coefficients at each offset location
            avals = np.interp(rr, xv, im_mask[nx//2,:]**2)
            bvals = 1 - avals

            # Linearly combine PSFs
            fov_pix = self.psf_info['fov_pix']
            psf_center  = krebin(self.psf_center_over, (fov_pix,fov_pix))
            psf_offaxis = krebin(self.psf_offaxis_over, (fov_pix,fov_pix))
            psf_max = np.array([np.max(psf_offaxis*a + psf_center*b) 
                                for a,b in zip(avals,bvals)])
            psf_max[rr>10] = psf_max[rr<10].max()

        elif self.mask[-1]=='B': # Bar masks
            fov_asec = np.max([xpix,ypix]) * pixscale

            # Image mask
            im_mask = coron_trans(self.mask, fov=fov_asec, pixscale=pixscale, nd_squares=False)
            im_mask = pad_or_cut_to_size(im_mask, data_rebin.shape)
            
            # For offaxis PSF max values, use fiducial at bar_offset location
            bar_offset = self.bar_offset
            ny, nx = im_mask.shape
            xloc = int(bar_offset / pixscale + nx/2)
            mask_cut = im_mask[:,xloc]

            # Interpolate along the horizontal cut
            yv = (np.arange(ny) - ny/2) * pixscale
            avals = np.interp(rr, yv, mask_cut**2)
            bvals = 1 - avals

            # Get PSF at middle of bar
            vals = self.psf_center_offsets
            arr = np.array(self.psf_center_over)
            func = interp1d(vals, arr, axis=0, kind='linear')
            psf_center = func(bar_offset)

            # Linearly combine PSFs
            fov_pix = self.psf_info['fov_pix']
            psf_center  = krebin(psf_center, (fov_pix,fov_pix))
            psf_offaxis = krebin(self.psf_offaxis_over, (fov_pix,fov_pix))
            psf_max = np.array([np.max(psf_offaxis*a + psf_center*b) 
                                for a,b in zip(avals,bvals)])
            psf_max[rr>10] = psf_max[rr<10].max()

        
        # We also want to know the Poisson noise for the PSF values.
        # For instance, even if psf_max is significantly above, the
        # background noise, Poisson noise could conspire to reduce
        # the peak PSF value below the noise level.
        
        # Count rate necessary to obtain some nsig
        texp  = self.multiaccum_times['t_exp']
        p     = 1 / texp
        crate = (p*nsig**2 + nsig * np.sqrt((p*nsig)**2 + 4*stds**2)) / 2
        # Get total count rate
        crate /= psf_max
        
        # Compute contrast
        contrast = crate / self.star_flux()

        # Magnitude sensitivity
        star_mag = self.star_flux('vegamag')
        sen_mag = star_mag - 2.5*np.log10(contrast)

        return (rr, contrast, sen_mag)
        
    
    def gen_slope_image(self, PA=0, exclude_disk=False, exclude_planets=False, 
        exclude_noise=False, zfact=None, do_ref=False, im_star=None, **kwargs):
        """Create slope image of observation

        Parameters
        ----------
        PA : float
            Position angle of roll position (clockwise, from East to West)
        do_ref : bool
            Slope image for reference star observation defined in `self.nrc_ref`.
        exclude_disk : bool
            Do not include disk in final image (for radial contrast),
            but still add Poisson noise from disk.
        exclude_planets : bool
            Do not include planets in final image (for radial contrast),
            but still add Poisson noise from disk.
        exclude_noise : bool
            Don't add random Gaussian noise (detector+photon)
        zfact : float
            Zodiacal background factor (default=2.5)
        im_star : ndarray or None
            Pass a precomputed slope image of the stellar source already
            positioned at it's correct location.
            
        Keyword Args
        ------------
        quick_PSF : bool
            Rather than generate a spectrum-weighted PSF for planets, use 
            the cached PSF scaled by the photon count through the bandpass.
            Resulting PSFs are slightly less accurate, but much faster.
            Default is True.
        use_cmask : bool
            Use the coronagraphic mask image to attenuate planet or disk that
            is obscurred by a corongraphic mask feature.
        locstr : 
            Object name or RA/DEC (decimal degrees or sexigesimal)
        year : int
            Year of observation
        day : float
            Day of observation

        
        """

        if do_ref:
            obs = self.nrc_ref
            sp = self.sp_ref
        else:
            obs = self
            sp = self.sp_sci

        det = obs.Detectors[0]

        # Final image shape
        image_shape = ypix, xpix = (det.ypix, det.xpix)

        # Center location of star
        bar_offpix = self.bar_offset / self.pixelscale
        if ('FULL' in self.det_info['wind_mode']) and (self.mask is not None):
            cdict = coron_ap_locs(self.module, self.channel, self.mask, full=True)
            xcen, ycen = cdict['cen_V23']
            xcen += bar_offpix
        else:
            xcen, ycen = (xpix/2 + bar_offpix, ypix/2)
        delx, dely = (xcen - xpix/2, ycen - ypix/2)

        # Stellar PSF doesn't rotate
        if im_star is None:
            im_star = obs.gen_psf(sp, return_oversample=False)
            im_star = pad_or_cut_to_size(im_star, image_shape)
            im_star = fshift(im_star, delx=delx, dely=dely, pad=True)
        
        # Disk and Planet images
        if do_ref:
            no_disk = no_planets = True
        else:
            no_disk    = exclude_disk    and exclude_noise
            no_planets = exclude_planets and exclude_noise

        # Make sure to include planets and disks for Poisson noise calculations
        im_disk = 0 if no_disk    else obs.gen_disk_image(PA_offset=PA, **kwargs)
        im_pl   = 0 if no_planets else obs.gen_planets_image(PA_offset=PA, **kwargs)

        # Zodiacal bg levels
        fzodi = obs.bg_zodi(zfact, **kwargs)

        # Combine components
        im_final = im_star + im_disk + im_pl + fzodi

        # Noise per pixel
        if not exclude_noise:
            im_noise = det.pixel_noise(fsrc=im_final)
            # Add random noise
            im_final += np.random.normal(scale=im_noise)

        # Get rid of disk and planet emission
        # while keeping their noise contribution
        if exclude_disk: im_final -= im_disk
        if exclude_planets: im_final -= im_pl
        
        return im_final
        
        

        
    def saturation_levels(self, ngroup=2, do_ref=False, image=None, **kwargs):
        """Saturation levels.
        
        Create image showing level of saturation for each pixel.
        Saturation at different number of groups is possible with
        ngroup keyword. Returns an array the same shape as `det_info`.
        
        Parameters
        ----------
        ngroup : int
            How many group times to determine saturation level?
            If this number is higher than the total groups in ramp, 
            then a warning is produced. The default is ngroup=2, 
            A value of 0 corresponds to the so-called "zero-frame," 
            which is the very first frame that is read-out and saved 
            separately. This is the equivalent to ngroup=1 for RAPID
            and BRIGHT1 observations.
        do_ref : bool
            Get saturation levels for reference soure instead of science
        image : ndarray
            Rather than generating an image on the fly, pass a pre-computed
            slope image.

        Keyword Args
        ------------
        quick_PSF : bool
            Rather than generate a spectrum-weighted PSF for planets, use 
            the cached PSF scaled by the photon count through the bandpass.
            Resulting PSFs are slightly less accurate, but much faster.
            Default is True.
        use_cmask : bool
            Use the coronagraphic mask image to attenuate planet or disk that
            is obscurred by a corongraphic mask feature.
        zfact : float
            Zodiacal background factor (default=2.5)
        locstr : 
            Object name or RA/DEC (decimal degrees or sexigesimal)
        year : int
            Year of observation
        day : float
            Day of observation

        
        """
        
        assert ngroup >= 0
        
        if do_ref:
            obs = self.nrc_ref
            sp = self.sp_ref
        else:
            obs = self
            sp = self.sp_sci
        
        if ngroup > obs.det_info['ngroup']:
            _log.warning("Specified ngroup is greater than self.det_info['ngroup'].")

        t_frame = obs.multiaccum_times['t_frame']
        t_int = obs.multiaccum_times['t_int']
        if ngroup==0:
            t_sat = t_frame
        else:
            ma = obs.multiaccum
            nf = ma.nf; nd1 = ma.nd1; nd2 = ma.nd2
            t_sat = (nd1 + ngroup*nf + (ngroup-1)*nd2) * t_frame
        
        #if t_sat>t_int:
        #    _log.warning('ngroup*t_group is greater than t_int.')
    
        # Slope image of input source
        if image is None:
            image = self.gen_slope_image(do_ref=do_ref, exclude_noise=True, **kwargs)
        
        # Well levels after "saturation time"
        sat_level = image * t_sat / obs.well_level
    
        return sat_level


class obs_coronagraphy(obs_hci):
    """Deprecated class. Use :class:`obs_hci` instead."""
    # Turns out the paper is Spiegel & Burrows (2012), not 2011
    def __init__(self, *args, **kwargs):
                 
        _log.warning('obs_coronagraphy is depcrecated. Use obs_hci instead.')
        obs_hci.__init__(self, *args, **kwargs)



def model_to_hdulist(args_model, sp_star, filter_or_bp, 
    pupil=None, mask=None, module=None):

    """HDUList from model FITS file.
    
    Convert disk model to an HDUList with units of photons/sec/pixel.
    If observed filter is different than input filter, we assume that
    the disk has a flat scattering, meaning it scales with stellar
    continuum. Pixel sizes and distances are left unchanged, and 
    stored in header.
    
    Parameters
    ----------
    args_model - tuple
        Arguments describing the necessary model information:
            - fname   : Name of model file
            - scale0  : Pixel scale (in arcsec/pixel)
            - dist0   : Assumed model distance
            - wave_um : Wavelength of observation
            - units0  : Assumed flux units (ie., MJy/arcsec^2 or muJy/pixel)
    sp_star : :mod:`pysynphot.spectrum`
        A pysynphot spectrum of central star. Used to adjust observed
        photon flux if filter differs from model input
    filter_or_bp : str, :mod:`pysynphot.obsbandpass`
        Either the name of a filter or a Pysynphot bandpass.
    pupil : str, None
        Instrument pupil setting (Lyot mask, grism, DHS, etc.)
    mask : str, None
        Coronagraphic mask.
    module : str
        NIRCam module 'A' or 'B'.
    """

    #filt, mask, pupil = args_inst
    fname, scale0, dist0, wave_um, units0 = args_model
    wave0 = wave_um * 1e4
    
    # Get the bandpass for the filter, pupil, and mask
    # This corresponds to the flux at the entrance pupil
    # for the particular filter.
    #bp = read_filter(filter, pupil=pupil, mask=mask, module=module)
    # Get filter throughput and create bandpass 
    if isinstance(filter_or_bp, six.string_types):
        filter = filter_or_bp
        bp = read_filter(filter, pupil=pupil, mask=mask, module=module)
    else:
        bp = filter_or_bp
        filter = bp.name



    # Detector pixel scale and PSF oversample
    #detscale = channel_select(bp)[0]
    #oversample = 4
    #pixscale_over = detscale / oversample
    
    #### Read in the image, then convert from mJy/arcsec^2 to photons/sec/pixel

    # Open file
    hdulist = fits.open(fname)
    #data    = hdulist[0].data#.copy()
    #header  = hdulist[0].header
    #hdutemp.close()

    # Break apart units0
    units_list = units0.split('/')
    if 'Jy' in units_list[0]:
        units_pysyn = S.units.Jy()
    if 'mJy' in units_list[0]:
        units_pysyn = S.units.mJy()
    if 'muJy' in units_list[0]:
        units_pysyn = S.units.muJy()
    if 'nJy' in units_list[0]:
        units_pysyn = S.units.nJy()    
    if 'MJy' in units_list[0]:
        hdulist[0].data *= 1000 # Convert to Jy
        units_pysyn = S.units.Jy()

    # Convert from input units to photlam (photons/sec/cm^2/A/angular size)
    # Compare observed wavelength to image wavelength
    wave_obs = bp.avgwave() # Current bandpass wavelength
    im = units_pysyn.ToPhotlam(wave0, hdulist[0].data)

    # We want to assume scattering is flat in photons/sec/A
    # This means everything scales with stellar continuum
    sp_star.convert('photlam') 
    im *= sp_star.sample(wave_obs) / sp_star.sample(wave0)

    # Convert to photons/sec/pixel
    im *= bp.equivwidth() * S.refs.PRIMARY_AREA
    # If input units are per arcsec^2 then scale by pixel scale
    # This will be ph/sec for each oversampled pixel
    if ('arcsec' in units_list[1]) or ('asec' in units_list[1]):
        im *= scale0**2
    elif 'mas' in units_list[1]:
        im *= (scale0*1000)**2
        
    # Save into HDUList
    hdulist[0].data = im
        
    hdulist[0].header['UNITS']    = 'photons/sec'
    hdulist[0].header['PIXELSCL'] = (scale0, 'arcsec/pixel')
    hdulist[0].header['DISTANCE'] = (dist0, 'parsecs')

    return hdulist


def _wrap_convolve_for_mp(args):
    """
    Internal helper routine for parallelizing computations across multiple processors.

    Create a list of arguments to pass to this function:
        worker_args = [(inst, image, rho, offset_list, i) for i,inst in enumerate(nrc_star_list)]

    Then create a theadpool:
        pool = mp.Pool(nproc)
        images = pool.map(_wrap_coeff_for_mp, worker_args)
        pool.close()
        images = np.array(images)

    For single processing, just use:
        images = [_wrap_convolve_for_mp(wa) for wa in worker_args]
        images = np.array(images)

    For multiprocessing:
    """

    psf, model, rho, offset_list, i = args

    noff = len(offset_list)
    if noff==1:
        r1 = 0; r2 = rho.max()+1	
    elif i==0:
        r1 = offset_list[i]
        r2 = (offset_list[i] + offset_list[i+1]) / 2.
    elif i==noff-1:
        r1 = (offset_list[i] + offset_list[i-1]) / 2.
        r2 = rho.max()+1.
    else:
        r1 = (offset_list[i] + offset_list[i-1]) / 2.
        r2 = (offset_list[i] + offset_list[i+1]) / 2.

    ind = (rho>=r1) & (rho<r2)
    im_temp = model.copy()
    im_temp[~ind] = 0
    
    # Normalize PSF sum to 1.0
    # Otherwise convolve_fft may throw an error if psf.sum() is too small
    norm = psf.sum()
    psf = psf / norm
    res = convolve_fft(im_temp, psf, fftn=fftpack.fftn, ifftn=fftpack.ifftn, allow_huge=True)
    res *= norm
    
    return res

def _wrap_conv_trans_for_mp(args):
    """
    Internal helper routine for parallelizing computations across multiple processors.

    Create a list of arguments to pass to this function:
        worker_args = [(inst, image, rho, offset_list, i) for i,inst in enumerate(nrc_star_list)]

    Then create a theadpool:
        pool = mp.Pool(nproc)
        images = pool.map(_wrap_coeff_for_mp, worker_args)
        pool.close()
        images = np.array(images)

    For single processing, just use:
        images = [_wrap_convolve_for_mp(wa) for wa in worker_args]
        images = np.array(images)

    For multiprocessing:
    """

    psf, model, tvals_edges, cmask, i = args
    
    tvals1 = tvals_edges[i]
    tvals2 = tvals_edges[i+1]
    ind = (cmask>=tvals1) & (cmask<=tvals2)

    im_temp = model.copy()
    im_temp[~ind] = 0
    
    # Normalize PSF sum to 1.0
    # Otherwise convolve_fft may throw an error if psf.sum() is too small
    norm = psf.sum()
    psf = psf / norm
    res = convolve_fft(im_temp, psf, fftn=fftpack.fftn, ifftn=fftpack.ifftn, allow_huge=True)
    res *= norm
    
    return res



def plot_contrasts(curves, nsig, wfe_list, obs=None, ax=None, 
    colors=None, xr=[0,10], yr=[25,5], return_axes=False):
    """Plot contrast curves
    
    Plot a series of contrast curves for corresponding WFE drifts.
    
    Parameters
    ----------
    curves : list
        A list with length corresponding to `wfe_list`. Each list element
        has three arrays in a tuple: the radius in arcsec, n-sigma contrast,
        and n-sigma sensitivity limit (vega mag). 
    nsig : float
        N-sigma limit corresponding to sensitivities/contrasts.
    wfe_list : array-like
        List of WFE drift values corresponding to each set of sensitivities
        in `curves` argument.
        
    Keyword Args
    ------------
    obs : :class:`obs_hci`
        Corresponding observation class that created the contrast curves.
        Uses distances and stellar magnitude to plot contrast and AU
        distances on opposing axes.
    ax : matplotlib.axes
        Axes on which to plot curves.
    colors : None, array-like
        List of colors for contrast curves. Default is gradient of blues.
    return_axes : bool
        Return the matplotlib axes to continue plotting. If `obs` is set,
        then this returns three sets of axes.
    
    """
    if ax is None:
        fig, ax = plt.subplots()
    if colors is None:
        lin_vals = np.linspace(0.2,0.7,len(wfe_list))
        colors = plt.cm.Blues_r(lin_vals)[::-1]
    for j, wfe_ref_drift in enumerate(wfe_list):
        rr, contrast, mag_sens = curves[j]
        label='$\Delta$' + "WFE = {} nm".format(wfe_list[j])
        ax.plot(rr, mag_sens, label=label, color=colors[j], zorder=1, lw=2)

    if xr is not None: ax.set_xlim(xr)
    if yr is not None: ax.set_ylim(yr)

        
    ax.xaxis.get_major_locator().set_params(nbins=9, steps=[1, 2, 5, 10])
    ax.yaxis.get_major_locator().set_params(nbins=9, steps=[1, 2, 5, 10])

    ax.set_ylabel('{:.0f}-$\sigma$ Sensitivities (mag)'.format(nsig))
    ax.set_xlabel('Separation (arcsec)')

    # Plot opposing axes in alternate units
    if obs is not None:
        yr1 = np.array(ax.get_ylim())
        yr2 = 10**((obs.star_flux('vegamag') - yr1) / 2.5)
        ax2 = ax.twinx()
        ax2.set_yscale('log')
        ax2.set_ylim(yr2)
        ax2.set_ylabel('{:.0f}-$\sigma$ Contrast'.format(nsig))

        ax3 = ax.twiny()
        xr3 = np.array(ax.get_xlim()) * obs.distance
        ax3.set_xlim(xr3)
        ax3.set_xlabel('Separtion (AU)')
        
        ax3.xaxis.get_major_locator().set_params(nbins=9, steps=[1, 2, 5, 10])
        
        if return_axes: 
            return (ax, ax2, ax3)
    else:
        if return_axes: 
            return ax
        

def planet_mags(obs, age=10, entropy=13, mass_list=[10,5,2,1], av_vals=[0,25], atmo='hy3s', **kwargs):
    """Determine series of exoplanet magnitudes for given observation."""

    if av_vals is None:
        av_vals = [0,0]
        
    pmag = {}
    for i,m in enumerate(mass_list):
        flux_list = []
        for j,av in enumerate(av_vals):
            sp = obs.planet_spec(mass=m, age=age, Av=av, entropy=entropy, atmo=atmo, **kwargs)
            sp_obs = S.Observation(sp, obs.bandpass, binset=obs.bandpass.wave)
            flux = sp_obs.effstim('vegamag')
            flux_list.append(flux)
        pmag[m] = tuple(flux_list)
        
    return pmag

import matplotlib.patches as patches
def plot_planet_patches(ax, obs, age=10, entropy=13, mass_list=[10,5,2,1], av_vals=[0,25], **kwargs):
    """Plot exoplanet magnitudes in region corresponding to extinction values."""
    
    xlim = ax.get_xlim()
    cols = plt.cm.tab10(np.linspace(0,1,10))
    
    #lin_vals = np.linspace(0,0.5,4)
    #cols = plt.cm.Purples_r(lin_vals)[::-1]
    
    dist = obs.distance
    
    if entropy<8: entropy=8
    if entropy>13: entropy=13

    pmag = planet_mags(obs, age, entropy, mass_list, av_vals, **kwargs)
    for i,m in enumerate(mass_list):
        label = 'Mass = {} '.format(m) + '$M_{\mathrm{Jup}}$'
        if av_vals is None:
            ax.plot(xlim, pmag[m], color=cols[i], lw=1, ls='--', label=label)        
        else:
            pm_min, pm_max = pmag[m]
            rect = patches.Rectangle((xlim[0], pm_min), xlim[1], pm_max-pm_min, 
                                     alpha=0.2, color=cols[i], label=label, zorder=2)
            ax.add_patch(rect)
            ax.plot(xlim, [pm_min]*2, color=cols[i], lw=1, alpha=0.3)
            ax.plot(xlim, [pm_max]*2, color=cols[i], lw=1, alpha=0.3)
        
    entropy_switch = {13:'Hot', 8:'Cold'}
    entropy_string = entropy_switch.get(entropy, "Warm")
    ent_str = entropy_string + ' Start'
    
    if av_vals is None:
        av_str = ''
    else:
        av_str = ' ($A_V = [{:.0f},{:.0f}]$)'.format(av_vals[0],av_vals[1])
    #age_str = 'Age = {:.0f} Myr; '.format(age)
    #dist_str = 'Dist = {:.1f} pc; '.format(dist) if dist is not None else ''
    #dist_str=""

    #ax.set_title('{} -- {} ({}{}{})'.format(obs.filter,ent_str,age_str,dist_str,av_str))

    #ax.set_title('{} -- {}{}'.format(obs.filter,ent_str,av_str))


def plot_hdulist(hdulist, xr=None, yr=None, ax=None, return_ax=False, 
    cmap=None, scale='linear', vmin=None, vmax=None, axes_color='white'):

    from webbpsf import display_psf

    if ax is None:
        fig, ax = plt.subplots()
    if cmap is None:
        cmap = matplotlib.rcParams['image.cmap']
    
    data = hdulist[0].data
    if vmax is None:
        vmax = 0.75 * data.max() if scale=='linear' else vmax
    if vmin is None:
        vmin = 0 if scale=='linear' else vmax/1e6

    ax, cb = display_psf(hdulist, ax=ax, title='', colorbar=True, cmap=cmap,
                         scale=scale, vmin=vmin, vmax=vmax, return_ax=True)
    cb.set_label('Counts per pixel')
    
    ax.set_xlim(xr)
    ax.set_ylim(yr)
    ax.set_xlabel('Arcsec')
    ax.set_ylabel('Arcsec')


    ax.tick_params(axis='both', color=axes_color, which='both')
    for k in ax.spines.keys():
        ax.spines[k].set_color(axes_color)

    ax.xaxis.get_major_locator().set_params(nbins=9, steps=[1, 2, 5, 10])
    ax.yaxis.get_major_locator().set_params(nbins=9, steps=[1, 2, 5, 10])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

root = path.abspath(path.dirname(__file__))

from pynrc.version import __version__
version = __version__

import os, sys

if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    print("You probably want to also tag the version now:")
    print("  python setup.py tag")
    sys.exit()

# Requires .pypirc
#[pypitest]
#repository: https://test.pypi.org/legacy/
#username: jarron
#password: ****
if sys.argv[-1] == 'pubtest':
    os.system("python setup.py sdist upload -r pypitest")
    os.system("python setup.py bdist_wheel upload -r pypitest")
    sys.exit()
    
if sys.argv[-1] == 'tag':
    os.system("git tag -a v%s -m 'Release %s'" % (version, version))
    os.system("git push origin v%s" % (version))
    sys.exit()


if sys.argv[-1] == 'test':
    test_requirements = [
        'pytest',
        'coverage'
    ]
    try:
        modules = map(__import__, test_requirements)
    except ImportError as e:
        err_msg = e.message.replace("No module named ", "")
        msg = "%s is not installed. Install your test requirments." % err_msg
        raise ImportError(msg)
        
    print('pyNRC testing not yet implemented!!')
    os.system('py.test')
    sys.exit()

# Get the long description from the README and HISTORY files
with open('README.rst') as readme_file:
    readme = readme_file.read()
with open('HISTORY.rst') as history_file:
    history = history_file.read()

#requirements = ['Click>=6.0', ]
requirements = ['Click>=6.0',
          'numpy>=1.10.0',
          'matplotlib>=1.5.0',
          'scipy>=0.16.0',
          'astropy<3.0',
          'pysynphot>=0.9',
          'poppy>=0.6.1',
          'webbpsf>=0.6.0',
          'jwxml>=0.3.0',
      ]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]


setup(
    name='pynrc',

    # Versions should comply with PEP440.
    version=version,

    description="JWST NIRCam ETC and Simulator",
    long_description=readme + '\n\n' + history,

    # The project's main homepage.
    url='https://github.com/JarronL/pynrc',

    # Author details
    author='Jarron Leisenring',
    author_email='jarronl@email.arizona.edu',
    license='MIT license',
    keywords='jwst nircam etc simulator',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Astronomy',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    #packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    packages=find_packages(include=['pynrc']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    #extras_require={
    #    'dev': ['check-manifest>=0.34', 'lxml>=3.6.4', 'pytest>=3.0.2'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    #package_data={
    #    'pynrc': ['package_data.dat'],
    #},

    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False,
)


"""
pyNRC - Python ETC and Simulator for JWST NIRCam
----------------------------------------------------------------------------

pyNRC is a set of Python-based tools for planning observations with JWST NIRCam, 
such as an ETC, a simple image slope simulator, and an enhanced data simulator.
This package works for a variety of NIRCam observing modes including direct imaging, 
coronagraphic imaging, slitless grism spectroscopy, DHS observations, 
and weak lens imaging. All PSFs are generated via WebbPSF 
(https://webbpsf.readthedocs.io) to reproduce realistic JWST images and spectra.

Developed by Jarron Leisenring and contributors at University of Arizona (2015-18).
"""
from __future__ import absolute_import, division, print_function, unicode_literals


from .version import __version__

import astropy
from astropy import config as _config

import tempfile

class Conf(_config.ConfigNamespace):

    # Path to data files for pynrc. 
    # The environment variable $PYNRC_PATH takes priority.
    import os
    
    on_rtd = os.environ.get('READTHEDOCS') == 'True'
    
    if on_rtd:
        path = tempfile.gettempdir()
    else:
        path = os.getenv('PYNRC_PATH')
        if path is None:
            print("WARNING: Environment variable $PYNRC_PATH is not set!")
            print("  Setting PYNRC_PATH to temporary directory.")
            path = tempfile.gettempdir()
            print("  {}".format(path))
            #raise EnvironmentError("Environment variable $PYNRC_PATH is not set!")
        if not os.path.isdir(path):
            #print ("WARNING: PYNRC_PATH ({}) is not a valid directory path!".format(path))
            raise IOError("PYNRC_PATH ({}) is not a valid directory path!".format(path))
            
    if '/' not in path[-1]: 
        # Make sure there is a '/' at the end of the path name
        path = path + '/'

    PYNRC_PATH = _config.ConfigItem(path, 'Directory path to data files \
                                    required for pynrc calculations.')

    logging_level = _config.ConfigItem(
        ['INFO', 'DEBUG', 'WARN', 'WARNING', 'ERROR', 'CRITICAL', 'NONE'],
        'Desired logging level for pyNRC.'
    )
    default_logging_level = _config.ConfigItem('INFO', 
        'Logging verbosity: one of {DEBUG, INFO, WARN, ERROR, or CRITICAL}')
    logging_filename = _config.ConfigItem("none", "Desired filename to save log messages to.")
    logging_format_screen = _config.ConfigItem(
        '[%(name)10s:%(levelname)s] %(message)s', 'Format for lines logged to the screen.'
    )
    logging_format_file = _config.ConfigItem(
        '%(asctime)s [%(name)s:%(levelname)s] %(filename)s:%(lineno)d: %(message)s',
        'Format for lines logged to a file.'
    )

conf = Conf()


from .logging_utils import setup_logging#, restart_logging
setup_logging(conf.default_logging_level, verbose=False)

from .nrc_utils import (read_filter, pix_noise, nrc_header, bp_2mass, bp_wise, \
                        stellar_spectrum, source_spectrum)

from .pynrc_core import (multiaccum, DetectorOps, NIRCam, planets_sb11, planets_sb12)

from .obs_nircam import (obs_coronagraphy, obs_hci)

#from .ngNRC import slope_to_ramp, nproc_use_ng

from .maths import *

from .simul import ngNRC

from .reduce import ref_pixels


def _reload(name="pynrc"):
    """
    Simple reload function to test code changes without restarting python.
    There may be some weird consequences and bugs that show up, such as
    functions and attributes deleted from the code still stick around after
    the reload. Although, this is even true with ``importlib.reload(pynrc)``.

    Other possible ways to reload on-the-fly: 
       
    from importlib import reload
    reload(pynrc)

    # Delete classes/modules to reload
    import sys
    del sys.modules['pynrc.obs_nircam'] 
    """
    import imp
    imp.load_module(name,*imp.find_module(name))

    print("{} reloaded".format(name)) 


